###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.examples module of the BEAT platform.         #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import bob.learn.em


def gmm_from_data(data):
    """Unmangles a bob.learn.em.GMMMachine from a BEAT Data object"""

    dim_c, dim_d = data.means.shape
    gmm = bob.learn.em.GMMMachine(dim_c, dim_d)
    gmm.weights = data.weights
    gmm.means = data.means
    gmm.variances = data.variances
    gmm.variance_thresholds = data.variance_thresholds
    return gmm


def isvbase_from_data(data, ubm):
    """Unmangles a bob.learn.em.ISVBase from a BEAT Data object"""

    dim_cd, dim_u = data.subspace_u.shape
    isvbase = bob.learn.em.ISVBase(ubm, dim_u)
    isvbase.u = data.subspace_u
    isvbase.d = data.subspace_d
    return isvbase


def stats_from_data(data):
    """Unmangles a bob.learn.em.GMMStats from a BEAT Data object"""

    dim_c, dim_d = data.sum_px.shape
    stat = bob.learn.em.GMMStats(dim_c, dim_d)
    stat.t = long(data.t)
    stat.n = data.n
    stat.sum_px = data.sum_px
    stat.sum_pxx = data.sum_pxx
    return stat


def isvmachine_from_data(data, isvbase):
    """Unmangles a bob.learn.em.ISVBase from a BEAT Data object"""

    isvmachine = bob.learn.em.ISVMachine(isvbase)
    isvmachine.z = data.latent_z
    return isvmachine




class Algorithm:

    def __init__(self):
        self.ubm       = None
        self.isvbase   = None
        self.templates = None


    def process(self, inputs, outputs):

        # retrieve the UBM once
        if self.ubm is None:
            inputs['ubm'].next()
            self.ubm = gmm_from_data(inputs['ubm'].data)

        # retrieve the ISVBase once
        if self.isvbase is None:
            inputs['isvbase'].next()
            self.isvbase = isvbase_from_data(inputs['isvbase'].data, self.ubm)


        # retrieve all the templates once
        if self.templates is None:
            self.templates = {}
            group = inputs.groupOf('template_model')

            while group.hasMoreData():
                group.next()

                template_id = group['template_id'].data.value

                self.templates[template_id] = dict(
                    client_id = group['template_client_id'].data.value,
                    model = isvmachine_from_data(group['template_model'].data, self.isvbase),
                )

        # process the probe
        template_ids = inputs['template_ids'].data.value
        statistics = stats_from_data(inputs['probe_statistics'].data)
        ux = inputs['probe_isv_offset'].data.value

        scores = []
        for template_id in template_ids:
            template_client_identity = self.templates[template_id]['client_id']
            model = self.templates[template_id]['model']


            score = model.forward_ux(statistics, ux)

            scores.append({
                'template_identity': template_client_identity,
                'score': score,
            })

        outputs['scores'].write({
                'client_identity': inputs['probe_client_id'].data.value,
                'scores': scores
            },
        )

        return True
