###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.examples module of the BEAT platform.         #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import numpy
import bob.io.base
import bob.db.avspoof
import scipy.io.wavfile

from bob.db.avspoof.driver import Interface
INFO = Interface()
SQLITE_FILE = INFO.files()[0]


class RecognitionTraining:

    def setup(self, root_folder, outputs, parameters, force_start_index=None, force_end_index=None):
        self.root_folder = root_folder
        self.outputs     = outputs
        self.parameters  = parameters

        # verification.Database implements bob.db.verification.utils interface
#        self.db = bob.db.avspoof.verification.Database(original_directory=self.root_folder, original_extension='.wav')
        self.db = bob.db.avspoof.Database()

        # retrieve all file paths for the given parameters
        self.objs = sorted(self.db.objects(protocol=parameters['protocol'], groups=parameters['group'], cls=('enroll', 'probe')), key=lambda x: x.id)
        
#        print("protocol: %s, group: %s, objects: %d" % (parameters['protocol'], parameters['group'], len(self.objs)))

        if parameters['protocol'] == "smalltest":
            assert (len(self.objs) == 741)
        else:
            assert (len(self.objs) == 4973)
        
        self.next_index       = 0

        self.force_start_index = force_start_index
        self.force_end_index = force_end_index

        # Retrieve only 'useful' data, as ignore everything before force_start_index and after force_end_index
        ### Start index
        if self.force_start_index is not None:
            self.next_index = self.force_start_index
        ### End index
        if self.force_end_index is not None:
            self.objs = self.objs[:self.force_end_index+1]

        return True


    def done(self):
        return (self.next_index >= len(self.objs))


    def next(self):

        obj = self.objs[self.next_index]

        if self.outputs['file_id'].isConnected():
            self.outputs['file_id'].write({'text': obj.id}, self.next_index)

        if self.outputs['client_id'].isConnected():
            self.outputs['client_id'].write({'text': obj.client_id}, self.next_index)

        # read audio data
        if self.outputs['speech'].isConnected():
            rate, audio = scipy.io.wavfile.read(obj.make_path(directory=self.root_folder, extension='.wav'))
            data = {
                'value': numpy.cast['float'](audio)
            }
            self.outputs['speech'].write(data, self.next_index)

        self.next_index += 1

        return True


class RecognitionTemplates:
    # Reasoning: Each client may have a number of models in certain databases.
    # So, each model receives an unique identifier. Those identifiers are linked
    # to the client identifier and contain a number of samples to generate the
    # model from.

    def setup(self, root_folder, outputs, parameters, force_start_index=None, force_end_index=None):
        self.root_folder            = root_folder
        self.outputs                = outputs
        self.parameters             = parameters

        # verification.Database implements bob.db.verification.utils interface
#        self.db                     = bob.db.avspoof.verification.Database(original_directory=self.root_folder, original_extension='.wav')
        self.db = bob.db.avspoof.Database()

        # template Ids are model Ids (which again are just Ids of clients) in the terminoloy of the bob.db interfaces
        if parameters['protocol'] == "smalltest":
            if parameters['group'] == "train":
                self.template_ids = sorted([1, 3])
            if parameters['group'] == "devel":
                self.template_ids = sorted([15, 20])
            if parameters['group'] == "test":
                self.template_ids = sorted([18, 33])
        else:
            self.template_ids = sorted([client.id for client in self.db.clients(groups=parameters['group'])])
        
#        print("protocol: %s, group: %s, purpose: %s, template_ids: %d" % (parameters['protocol'], parameters['group'], parameters['purpose'], len(self.template_ids)))

        if parameters['protocol'] == "smalltest":
            assert (len(self.template_ids) == 2)
        else:
            if parameters['group'] == "test":
                assert (len(self.template_ids) == 16)
            else:
                assert (len(self.template_ids) == 14)
        
        self.objs = []
        # store the File objects for each template
        for template_id in self.template_ids:
          self.objs.append(sorted(self.db.objects(protocol=parameters['protocol'], groups=parameters['group'], cls=parameters['purpose'], clients = (template_id,)), key=lambda x: x.id))
        
        self.current_template_index = 0
        self.current_obj_index      = 0
        self.next_index             = 0

        self.force_start_index = force_start_index
        self.force_end_index   = force_end_index

        # Retrieve only 'useful' data
        ### End index
        #if self.force_end_index is not None:
        #    self.objs = self.objs[:self.force_end_index+1]
        ### Start index
        if self.force_start_index is None:
            self.force_start_index = 0
        while self.next_index < self.force_start_index:
            self.next()

        return True


    def done(self):
        return (self.current_template_index >= len(self.template_ids)) or (self.force_end_index is not None and self.force_end_index < self.next_index)


    def next(self):
        # the length of the batch of File objects in the current template
        len_template_objs = len(self.objs[self.current_template_index])
        # current file object that we want to read and output
        obj = self.objs[self.current_template_index][self.current_obj_index]


        if self.current_obj_index == 0:
            # we update the 'template_id' and 'client_id' only when the current template/model changes
            if self.force_start_index <= self.next_index and (self.force_end_index is None or self.force_end_index >= self.next_index):
                # For this database, 'self.template_ids[self.current_template_index]' corresponds to 'obj.client_id'
                if self.outputs['template_id'].isConnected():
                    self.outputs['template_id'].write({'text': self.template_ids[self.current_template_index]}, self.next_index+len_template_objs-1)

                if self.outputs['client_id'].isConnected():
                    self.outputs['client_id'].write({'text': obj.client_id}, self.next_index+len_template_objs-1)

        if self.outputs['file_id'].isConnected() or self.outputs['speech'].isConnected():
            if self.force_start_index <= self.next_index and (self.force_end_index is None or self.force_end_index >= self.next_index):
                if self.outputs['file_id'].isConnected():
                    self.outputs['file_id'].write({'text': obj.id}, self.next_index)

                if self.outputs['speech'].isConnected():
                    rate, audio = scipy.io.wavfile.read(obj.make_path(directory=self.root_folder, extension='.wav'))
                    data = {
                        'value': numpy.cast['float'](audio)
                    }
                    self.outputs['speech'].write(data, self.next_index)

            self.next_index += 1
            self.current_obj_index += 1
        else:
            self.next_index += len_template_objs
            self.current_obj_index = len_template_objs

        if self.current_obj_index == len_template_objs:
            self.current_obj_index = 0
            self.current_template_index += 1

        return True


class Probes:
    # Reasoning: Each client may have a number of probes. Each probe may be
    # composed of any number of sample. So, each probe receives an unique
    # identifier. Those identifiers are linked to the client identifier and
    # contain a number of images to generated the match/comparison score. Each
    # probe must be matched against a number of models defined by a list of
    # client identifiers.

    def setup(self, root_folder, outputs, parameters, force_start_index=None, force_end_index=None):
        self.root_folder = root_folder
        self.outputs     = outputs
        self.parameters  = parameters

        # verification.Database implements bob.db.verification.utils interface
#        self.db                     = bob.db.avspoof.verification.Database(original_directory=self.root_folder, original_extension='.wav')
        self.db = bob.db.avspoof.Database()

        # template Ids are model Ids (which again are just Ids of clients) in the terminoloy of the bob.db interfaces
        if parameters['purpose'] == "attack":
            self.template_ids = None
        else:
            if parameters['protocol'] == "smalltest":
                if parameters['group'] == "train":
                    self.template_ids = sorted([1, 3])
                if parameters['group'] == "devel":
                    self.template_ids = sorted([15, 20])
                if parameters['group'] == "test":
                    self.template_ids = sorted([18, 33])
            else:
                self.template_ids = sorted([client.id for client in self.db.clients(groups=parameters['group'])])
        
        self.objs                   = sorted(self.db.objects(protocol=parameters['protocol'], groups=parameters['group'], cls=parameters['purpose']), key=lambda x: x.id)

#        print("protocol: %s, group: %s, purpose: %s, objects: %d" % (parameters['protocol'], parameters['group'], parameters['purpose'], len(self.objs)))

        if parameters['purpose'] == "attack":
            if parameters['protocol'] == "smalltest":
                assert (len(self.objs) == 6910)
            if parameters['protocol'] == "grandtest":
                if parameters['group'] == "train":
                    assert (len(self.objs) == 56470)
                if parameters['group'] == "devel":
                    assert (len(self.objs) == 56470)
                if parameters['group'] == "test":
                    assert (len(self.objs) == 63380)
            if parameters['protocol'] == "physical_access":
                if parameters['group'] == "train":
                    assert (len(self.objs) == 38580)
                if parameters['group'] == "devel":
                    assert (len(self.objs) == 38580)
                if parameters['group'] == "test":
                    assert (len(self.objs) == 43320)
            if parameters['protocol'] == "logical_access":
                if parameters['group'] == "train":
                    assert (len(self.objs) == 17890)
                if parameters['group'] == "devel":
                    assert (len(self.objs) == 17890)
                if parameters['group'] == "test":
                    assert (len(self.objs) == 20060)
        else:
            if parameters['protocol'] == "smalltest":
                assert (len(self.template_ids) == 2)
            else:
                if parameters['group'] == "test":
                    assert (len(self.template_ids) == 16)
                else:
                    assert (len(self.template_ids) == 14)
            if parameters['protocol'] == "smalltest":
                if parameters['group'] == "train":
                    assert (len(self.objs) == 623)
                if parameters['group'] == "devel":
                    assert (len(self.objs) == 586)
                if parameters['group'] == "test":
                    assert (len(self.objs) == 573)
            else:
                if parameters['group'] == "train":
                    assert (len(self.objs) == 4193)
                if parameters['group'] == "devel":
                    assert (len(self.objs) == 4225)
                if parameters['group'] == "test":
                    assert (len(self.objs) == 4708)

        
        self.next_index  = 0

        self.force_start_index = force_start_index
        self.force_end_index = force_end_index

        # Retrieve only 'useful' data
        ### Start index
        if self.force_start_index is not None:
            self.next_index = self.force_start_index
        ### End index
        if self.force_end_index is not None:
            if self.force_end_index < len(self.objs):
                self.objs = self.objs[:self.force_end_index+1]

        return True


    def done(self):
        return (self.next_index >= len(self.objs))


    def next(self):

        obj = self.objs[self.next_index]

        if self.outputs['file_id'].isConnected():
            self.outputs['file_id'].write({'text': obj.id}, self.next_index)

        if self.template_ids is None:  # we are processing attacks
            if self.outputs['attack_id'].isConnected():
                self.outputs['attack_id'].write({'text': obj.id}, self.next_index)
        else:  # we are porcessing real data
            if self.outputs['probe_id'].isConnected():
                self.outputs['probe_id'].write({'text': obj.id}, self.next_index)

        if self.outputs['client_id'].isConnected():
            self.outputs['client_id'].write({'text': obj.client_id}, self.next_index)

        if self.outputs['template_ids'].isConnected():
            if self.template_ids is None:  # it means we are processing attacks, so template_ids is just one client_id
                data = {
                  'text': numpy.array([obj.client_id], dtype=numpy.string_)
                }
            else:
                data = {
                  'text': numpy.array(self.template_ids, dtype=numpy.string_)
                }
            self.outputs['template_ids'].write(data, self.next_index)

        # read audio data
        if self.outputs['speech'].isConnected():
            rate, audio = scipy.io.wavfile.read(obj.make_path(directory=self.root_folder, extension='.wav'))
            data = {
                'value': numpy.cast['float'](audio)
            }
            self.outputs['speech'].write(data, self.next_index)

        self.next_index += 1

        return True


class SimpleAntispoofing:

    def setup(self, root_folder, outputs, parameters, force_start_index=None, force_end_index=None):
        self.root_folder = root_folder
        self.outputs     = outputs
        self.parameters  = parameters

        # By right we should take the bob.db.avspoof.spoofing.Database(), which implements antispoofing.utils.db interface, but that interface does not support Id's for file objects (see antispoofing.utils.File). It's a problem for us here, since we need to sort files by their Id's and also we need to yeild the id of each file.
        # So, we are using the direct access to the Database here.
        self.db = bob.db.avspoof.Database()

        # retrieve all file paths for the given parameters
        self.objs_real = sorted(self.db.objects(protocol=parameters['protocol'], groups=parameters['group'], cls='real'), key=lambda x: x.id)
#        print("protocol: %s, group: %s, real objects: %d" % (parameters['protocol'], parameters['group'], len(self.objs_real)))
        self.objs_attack = sorted(self.db.objects(protocol=parameters['protocol'], groups=parameters['group'], cls='attack'), key=lambda x: x.id)
#        print("protocol: %s, group: %s, attack objects: %d" % (parameters['protocol'], parameters['group'], len(self.objs_attack)))

        if parameters['protocol'] == "smalltest":
            assert (len(self.objs_attack) == 6910)
            if parameters['group'] == "train":
                assert (len(self.objs_real) == 741)
            if parameters['group'] == "devel":
                assert (len(self.objs_real) == 695)
            if parameters['group'] == "test":
                assert (len(self.objs_real) == 680)
        else:
            if parameters['group'] == "train":
                assert (len(self.objs_real) == 4973)
            if parameters['group'] == "devel":
                assert (len(self.objs_real) == 4995)
            if parameters['group'] == "test":
                assert (len(self.objs_real) == 5576)
            if parameters['protocol'] == "grandtest":
                if parameters['group'] == "train":
                    assert (len(self.objs_attack) == 56470)
                if parameters['group'] == "devel":
                    assert (len(self.objs_attack) == 56470)
                if parameters['group'] == "test":
                    assert (len(self.objs_attack) == 63380)
            if parameters['protocol'] == "physical_access":
                if parameters['group'] == "train":
                    assert (len(self.objs_attack) == 38580)
                if parameters['group'] == "devel":
                    assert (len(self.objs_attack) == 38580)
                if parameters['group'] == "test":
                    assert (len(self.objs_attack) == 43320)
            if parameters['protocol'] == "logical_access":
                if parameters['group'] == "train":
                    assert (len(self.objs_attack) == 17890)
                if parameters['group'] == "devel":
                    assert (len(self.objs_attack) == 17890)
                if parameters['group'] == "test":
                    assert (len(self.objs_attack) == 20060)

        # we have two sets of data, so to serve them sequentially, we need to know their sizes and
        # when the second set is following after the first set, hence, this approach with offsets
        self.offset_real    = 0
        self.offset_attack      = self.offset_real + len(self.objs_real)
        self.next_index       = 0

        self.force_start_index = force_start_index
        self.force_end_index = force_end_index

        # Retrieve only 'useful' data, as ignore everything before force_start_index and after force_end_index
        ### Start index
        if self.force_start_index is not None:
            self.next_index = self.force_start_index
        ### End index
        if self.force_end_index is not None:
            if self.force_end_index < self.offset_attack:  # it means we are still within real objects
                self.objs_real = self.objs_real[:self.force_end_index+1]
                self.offset_attack = []
            elif self.force_end_index < self.offset_attack + len(self.objs_attack) - 1:
                self.objs_attack = self.objs_attack[:self.force_end_index+1-self.offset_attack]

        return True


    def done(self):
        return (self.next_index >= len(self.objs_real) + len(self.objs_attack))


    def next(self):

        cls = ''
        attack_type = ''
        obj = None
        # if we are inside set of attacks, we yeild an attack object
        if self.next_index >= self.offset_attack:
          obj = self.objs_attack[self.next_index - self.offset_attack]
          cls = 'attack'
          # get the attack and compute the type of the attack
          attack = obj.get_attack()
          attack_type = attack.attack_support + attack.attack_device
        else:
          obj = self.objs_real[self.next_index - self.offset_real]
          cls = 'real'
          attack_type = 'human'

        if self.outputs['file_id'].isConnected():
            self.outputs['file_id'].write({'text': obj.id}, self.next_index)

        if self.outputs['client_id'].isConnected():
            self.outputs['client_id'].write({'text': obj.client_id}, self.next_index)

        # read audio data
        if self.outputs['speech'].isConnected():
            rate, audio = scipy.io.wavfile.read(obj.make_path(directory=self.root_folder, extension='.wav'))
            data = {
                'value': numpy.cast['float'](audio)
            }
            self.outputs['speech'].write(data, self.next_index)

        if self.outputs['attack_type'].isConnected():
            self.outputs['attack_type'].write({'text': attack_type}, self.next_index)

        if self.outputs['class'].isConnected():
            self.outputs['class'].write({'text': cls}, self.next_index)

        self.next_index += 1

        return True