###############################################################################
#                                                                             #
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.examples module of the BEAT platform.         #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import numpy as np
import bob.io.base
import bob.db.avspoof
import scipy.io.wavfile

from bob.db.avspoof.driver import Interface
INFO = Interface()
SQLITE_FILE = INFO.files()[0]


#----------------------------------------------------------


def get_client_end_index(objs, client_id, client_start_index,
                         start_index, end_index, indice_in_tuple=1):
    client_end_index = client_start_index

    while client_end_index + 1 <= end_index:
        obj = objs[client_end_index + 1 - start_index]

        if isinstance(obj, tuple):
            obj = obj[indice_in_tuple]

        if obj.client_id != client_id:
            return client_end_index

        client_end_index += 1

    return end_index


#----------------------------------------------------------


def get_value_end_index(objs, value, index_in_tuple, value_start_index,
                           start_index, end_index):
    value_end_index = value_start_index

    while value_end_index + 1 <= end_index:
        id = objs[value_end_index + 1 - start_index][index_in_tuple]

        if id != value:
            return value_end_index

        value_end_index += 1

    return end_index


#----------------------------------------------------------


class RecognitionTraining:
    """Outputs:
        - speech: "{{ system_user.username }}/array_1d_floats/1"
        - file_id: "{{ system_user.username }}/text/1"
        - client_id: "{{ system_user.username }}/text/1"

    One "file_id" is associated with a given "speech".
    Several "speech" are associated with a given "client_id".

    --------------- --------------- --------------- --------------- --------------- ---------------
    |    speech   | |    speech   | |    speech   | |    speech   | |    speech   | |    speech   |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   |
    --------------- --------------- --------------- --------------- --------------- ---------------
    ----------------------------------------------- -----------------------------------------------
    |                   client_id                 | |                   client_id                 |
    ----------------------------------------------- -----------------------------------------------
    """


    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = root_folder
        self.outputs = outputs

        # Open the database and load the objects to provide via the outputs
        self.db = bob.db.avspoof.Database()
        self.objs = sorted(self.db.objects(protocol=parameters['protocol'],
                                           groups=parameters['group'],
                                           cls=('enroll', 'probe')),
                           key=lambda x: (x.client_id, x.id))

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        obj = self.objs[self.next_index - self.start_index]

        # Output: client_id (only provide data when the client_id change)
        if self.outputs['client_id'].isConnected() and \
           self.outputs['client_id'].last_written_data_index < self.next_index:

            client_end_index = get_client_end_index(self.objs, obj.client_id,
                                                    self.next_index,
                                                    self.start_index,
                                                    self.end_index)

            self.outputs['client_id'].write(
                {
                    'text': str(obj.client_id)
                },
                client_end_index
            )

        # Output: file_id (provide data at each iteration)
        if self.outputs['file_id'].isConnected():
            self.outputs['file_id'].write(
                {
                    'text': str(obj.id)
                },
                self.next_index
            )

        # Output: speech (provide data at each iteration)
        if self.outputs['speech'].isConnected():
            rate, audio = scipy.io.wavfile.read(obj.make_path(directory=self.root_folder,
                                                              extension='.wav'))
            self.outputs['speech'].write(
                {
                    'value': np.cast['float'](audio)
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


class RecognitionTemplates:
    """Outputs:
        - speech: "{{ system_user.username }}/array_1d_floats/1"
        - file_id: "{{ system_user.username }}/text/1"
        - template_id: "{{ system_user.username }}/text/1"
        - client_id: "{{ system_user.username }}/text/1"

    One "file_id" is associated with a given "speech".
    Several "speech" are associated with a given "template_id".
    Several "template_id" are associated with a given "client_id".

    --------------- --------------- --------------- --------------- --------------- ---------------
    |    speech   | |    speech   | |    speech   | |    speech   | |    speech   | |    speech   |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   |
    --------------- --------------- --------------- --------------- --------------- ---------------
    ----------------------------------------------- -----------------------------------------------
    |                  template_id                | |                  template_id                |
    ----------------------------------------------- -----------------------------------------------
    -----------------------------------------------------------------------------------------------
    |                                         client_id                                           |
    -----------------------------------------------------------------------------------------------

    Note: for this particular database, there is only one "template_id"
    per "client_id".
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = root_folder
        self.outputs = outputs
        self.parameters = parameters

        # Open the database and load the objects to provide via the outputs
        self.db = bob.db.avspoof.Database()

        if parameters['protocol'] == "smalltest":
            if parameters['group'] == "train":
                template_ids = sorted([1, 3])
            if parameters['group'] == "devel":
                template_ids = sorted([15, 20])
            if parameters['group'] == "test":
                template_ids = sorted([18, 33])
        else:
            template_ids = [ client.id for client in self.db.clients(groups=self.parameters['group']) ]

        self.objs = []

        for template_id in template_ids:
            objs = self.db.objects(groups=self.parameters['group'],
                                   protocol=self.parameters['protocol'],
                                   cls=self.parameters['purpose'],
                                   clients=(template_id,))

            self.objs.extend([ (template_id, obj) for obj in objs ])

        self.objs = sorted(self.objs, key=lambda x: (x[1].client_id, x[0], x[1].id))

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        (template_id, obj) = self.objs[self.next_index - self.start_index]

        # Output: template_id (only provide data when the template_id change)
        if self.outputs['template_id'].isConnected() and \
           self.outputs['template_id'].last_written_data_index < self.next_index:

            template_end_index = get_value_end_index(self.objs, template_id, 0,
                                                     self.next_index,
                                                     self.start_index,
                                                     self.end_index)

            self.outputs['template_id'].write(
                {
                    'text': str(template_id)
                },
                template_end_index
            )


        # Output: client_id (only provide data when the client_id change)
        if self.outputs['client_id'].isConnected() and \
           self.outputs['client_id'].last_written_data_index < self.next_index:

            client_end_index = get_client_end_index(self.objs, obj.client_id,
                                                    self.next_index,
                                                    self.start_index,
                                                    self.end_index)

            self.outputs['client_id'].write(
                {
                    'text': str(obj.client_id)
                },
                client_end_index
            )

        # Output: file_id (provide data at each iteration)
        if self.outputs['file_id'].isConnected():
            self.outputs['file_id'].write(
                {
                    'text': str(obj.id)
                },
                self.next_index
            )

        # Output: speech (provide data at each iteration)
        if self.outputs['speech'].isConnected():
            rate, audio = scipy.io.wavfile.read(obj.make_path(directory=self.root_folder,
                                                              extension='.wav'))
            self.outputs['speech'].write(
                {
                    'value': np.cast['float'](audio)
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


class Probes:
    """Outputs:
        - speech: "{{ system_user.username }}/array_1d_floats/1"
        - file_id: "{{ system_user.username }}/text/1"
        - client_id: "{{ system_user.username }}/text/1"
        - template_ids: "{{ system_user.username }}/array_1d_text/1",

       And either:
        - probe_id: "{{ system_user.username }}/text/1",
       or:
        - attack_id: "{{ system_user.username }}/text/1",

    One "file_id" is associated with a given "speech".
    One "probe_id" is associated with a given "speech".
    Several "speech" are associated with a given "client_id".
    Several "client_id" are associated with a given "template_ids".

    Each probe must be matched against a number of templates defined by a list of
    client identifiers.

    --------------- --------------- --------------- --------------- --------------- ---------------
    |    speech   | |    speech   | |    speech   | |    speech   | |    speech   | |    speech   |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    |   probe_id  | |   probe_id  | |   probe_id  | |   probe_id  | |   probe_id  | |   probe_id  |
    --------------- --------------- --------------- --------------- --------------- ---------------
    ----------------------------------------------- -----------------------------------------------
    |                   client_id                 | |                   client_id                 |
    ----------------------------------------------- -----------------------------------------------
    -----------------------------------------------------------------------------------------------
    |                                        template_ids                                         |
    -----------------------------------------------------------------------------------------------
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = root_folder
        self.outputs = outputs
        self.parameters = parameters

        # Open the database and load the objects to provide via the outputs
        self.db = bob.db.avspoof.Database()

        if parameters['purpose'] == "attack":
            self.template_ids = None

        else:
            if parameters['protocol'] == "smalltest":
                if parameters['group'] == "train":
                    self.template_ids = sorted([1, 3])
                if parameters['group'] == "devel":
                    self.template_ids = sorted([15, 20])
                if parameters['group'] == "test":
                    self.template_ids = sorted([18, 33])
            else:
                self.template_ids = sorted([ client.id for client in self.db.clients(groups=self.parameters['group']) ])

        self.objs = sorted(self.db.objects(protocol=parameters['protocol'],
                                           groups=parameters['group'],
                                           cls=parameters['purpose']),
                           key=lambda x: (x.client_id, x.id))

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        obj = self.objs[self.next_index - self.start_index]

        if self.template_ids is None:

            # Output: template_ids (only provide data when the client_id change)
            if self.outputs['template_ids'].isConnected() and \
               self.outputs['template_ids'].last_written_data_index < self.next_index:

                client_end_index = get_client_end_index(self.objs, obj.client_id,
                                                        self.next_index,
                                                        self.start_index,
                                                        self.end_index)

                self.outputs['template_ids'].write(
                    {
                        'text': [ str(obj.client_id) ]
                    },
                    client_end_index
                )
        else:
            # Output: template_ids (only provide data once)
            if self.outputs['template_ids'].isConnected() and \
               self.outputs['template_ids'].last_written_data_index < self.next_index:

                self.outputs['template_ids'].write(
                    {
                        'text': [ str(x) for x in self.template_ids ]
                    },
                    self.end_index
                )


        # Output: client_id (only provide data when the client_id change)
        if self.outputs['client_id'].isConnected() and \
           self.outputs['client_id'].last_written_data_index < self.next_index:

            client_end_index = get_client_end_index(self.objs, obj.client_id,
                                                    self.next_index,
                                                    self.start_index,
                                                    self.end_index)

            self.outputs['client_id'].write(
                {
                    'text': str(obj.client_id)
                },
                client_end_index
            )

        # Output: probe_id (provide data at each iteration)
        if (self.outputs['probe_id'] is not None) and self.outputs['probe_id'].isConnected():
            self.outputs['probe_id'].write(
                {
                    'text': str(obj.id)
                },
                self.next_index
            )


        # Output: attack_id (provide data at each iteration)
        if (self.outputs['attack_id'] is not None) and self.outputs['attack_id'].isConnected():
            self.outputs['attack_id'].write(
                {
                    'text': str(obj.id)
                },
                self.next_index
            )


        # Output: file_id (provide data at each iteration)
        if self.outputs['file_id'].isConnected():
            self.outputs['file_id'].write(
                {
                    'text': str(obj.id)
                },
                self.next_index
            )

        # Output: speech (provide data at each iteration)
        if self.outputs['speech'].isConnected():
            rate, audio = scipy.io.wavfile.read(obj.make_path(directory=self.root_folder,
                                                              extension='.wav'))
            self.outputs['speech'].write(
                {
                    'value': np.cast['float'](audio)
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


class SimpleAntispoofing:
    """Outputs:
        - speech: "{{ system_user.username }}/array_1d_floats/1"
        - file_id: "{{ system_user.username }}/text/1"
        - client_id: "{{ system_user.username }}/text/1"
        - attack_type: "{{ system_user.username }}/text/1"
        - class: "{{ system_user.username }}/text/1"


    One "file_id" is associated with a given "speech".
    Several "speech" are associated with a given "client_id".
    Several "client_id" are associated with a given "template_id".

    --------------- --------------- --------------- --------------- --------------- ---------------
    |    speech   | |    speech   | |    speech   | |    speech   | |    speech   | |    speech   |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   |
    --------------- --------------- --------------- --------------- --------------- ---------------
    ------------------------------- ------------------------------- -------------------------------
    |          attack_type        | |          attack_type        | |          attack_type        |
    ------------------------------- ------------------------------- -------------------------------
    --------------------------------------------------------------- -------------------------------
    |                          client_id                          | |                   client_id
    --------------------------------------------------------------- -------------------------------
    -----------------------------------------------------------------------------------------------
    |                                          class                                              |
    -----------------------------------------------------------------------------------------------

    Note: for this particular database, there is only one "client_id"
    per "template_id".
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = root_folder
        self.outputs = outputs
        self.parameters = parameters

        # Open the database and load the objects to provide via the outputs
        self.db = bob.db.avspoof.Database()

        objs_real = sorted(self.db.objects(protocol=parameters['protocol'],
                                           groups=parameters['group'],
                                           cls='real'),
                           key=lambda x: (x.client_id, x.id))

        objs_attack = sorted(self.db.objects(protocol=parameters['protocol'],
                                             groups=parameters['group'],
                                             cls='attack'),
                             key=lambda x: (x.client_id, x.get_attack(), x.id))

        self.objs = [ ('real', 'human', x) for x in objs_real ] + \
                    [ ('attack', x.get_attack(), x) for x in objs_attack ]

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        (cls, attack_type, obj) = self.objs[self.next_index - self.start_index]

        # Output: class (only provide data when the class change)
        if self.outputs['class'].isConnected() and \
           self.outputs['class'].last_written_data_index < self.next_index:

            class_end_index = get_value_end_index(self.objs, cls, 0,
                                                  self.next_index,
                                                  self.start_index,
                                                  self.end_index)

            self.outputs['class'].write(
                {
                    'text': cls
                },
                class_end_index
            )


        # Output: attack_type (only provide data when the attack_type change)
        if self.outputs['attack_type'].isConnected() and \
           self.outputs['attack_type'].last_written_data_index < self.next_index:

            attack_type_end_index = get_value_end_index(self.objs, attack_type, 1,
                                                        self.next_index,
                                                        self.start_index,
                                                        self.end_index)

            self.outputs['attack_type'].write(
                {
                    'text': attack_type
                },
                attack_type_end_index
            )


        # Output: client_id (only provide data when the client_id change)
        if self.outputs['client_id'].isConnected() and \
           self.outputs['client_id'].last_written_data_index < self.next_index:

            client_end_index = get_client_end_index(self.objs, obj.client_id,
                                                    self.next_index,
                                                    self.start_index,
                                                    self.end_index,
                                                    indice_in_tuple=2)

            self.outputs['client_id'].write(
                {
                    'text': str(obj.client_id)
                },
                client_end_index
            )


        # Output: file_id (provide data at each iteration)
        if self.outputs['file_id'].isConnected():
            self.outputs['file_id'].write(
                {
                    'text': str(obj.id)
                },
                self.next_index
            )

        # Output: speech (provide data at each iteration)
        if self.outputs['speech'].isConnected():
            rate, audio = scipy.io.wavfile.read(obj.make_path(directory=self.root_folder,
                                                              extension='.wav'))
            self.outputs['speech'].write(
                {
                    'value': np.cast['float'](audio)
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


def setup_tests():
    # Install a mock read function for the audio files
    def mock_read(filename):
        return 44100, np.ndarray((128,))

    scipy.io.wavfile.read = mock_read


#----------------------------------------------------------


# Test the behavior of the views (on fake data)
if __name__ == '__main__':

    setup_tests()

    from beat.backend.python.database import DatabaseTester

    DatabaseTester('RecognitionTraining', RecognitionTraining,
        [
            'client_id',
            'file_id',
            'speech',
        ],
        dict(
            protocol = 'smalltest',
            group = 'train',
        ),
        irregular_outputs = [
            'client_id',
        ]
    )

    DatabaseTester('RecognitionTemplates', RecognitionTemplates,
        [
            'client_id',
            'template_id',
            'file_id',
            'speech',
        ],
        dict(
            protocol = 'smalltest',
            group = 'train',
            purpose = 'enroll',
        ),
        irregular_outputs = [
            'template_id',
            'client_id',
        ]
    )

    DatabaseTester('Probes (with probe_id)', Probes,
        [
            'template_ids',
            'client_id',
            'probe_id',
            'file_id',
            'speech',
        ],
        dict(
            protocol = 'smalltest',
            group = 'train',
            purpose = 'probe',
        ),
        irregular_outputs = [
            'client_id',
        ]
    )

    DatabaseTester('Probes (with attack_id)', Probes,
        [
            'template_ids',
            'client_id',
            'attack_id',
            'file_id',
            'speech',
        ],
        dict(
            protocol = 'smalltest',
            group = 'train',
            purpose = 'attack',
        ),
        irregular_outputs = [
            'template_ids',
            'client_id',
        ]
    )

    DatabaseTester('SimpleAntispoofing', SimpleAntispoofing,
        [
            'class',
            'client_id',
            'attack_type',
            'file_id',
            'speech',
        ],
        dict(
            protocol = 'smalltest',
            group = 'train',
        ),
        irregular_outputs = [
            'class',
            'client_id',
            'attack_type',
        ]
    )
