###############################################################################
#                                                                             #
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.examples module of the BEAT platform.         #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import os
import numpy as np
import bob.io.base
import bob.io.image
import bob.db.biowave_v1


#----------------------------------------------------------


def get_client_end_index(objs, client_id, client_start_index,
                         start_index, end_index):
    client_end_index = client_start_index

    while client_end_index + 1 <= end_index:
        obj = objs[client_end_index + 1 - start_index]

        if isinstance(obj, tuple):
            obj = obj[1]

        if obj.client_id != client_id:
            return client_end_index

        client_end_index += 1

    return end_index


#----------------------------------------------------------


def get_model_end_index(objs, model_id, model_start_index,
                        start_index, end_index):
    model_end_index = model_start_index

    while model_end_index + 1 <= end_index:
        obj = objs[model_end_index + 1 - start_index]

        if obj.model_id != model_id:
            return model_end_index

        model_end_index += 1

    return end_index


#----------------------------------------------------------


class View:
    """Outputs:
        - image: "{{ system_user.username }}/array_2d_uint8/1"
        - client_id: "{{ system_user.username }}/uint64/1"

    Several "image" are associated with a given "client_id".

    --------------- --------------- --------------- --------------- --------------- ---------------
    |    image    | |    image    | |    image    | |    image    | |    image    | |    image    |
    --------------- --------------- --------------- --------------- --------------- ---------------
    ----------------------------------------------- -----------------------------------------------
    |                   client_id                 | |                   client_id                 |
    ----------------------------------------------- -----------------------------------------------
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = root_folder
        self.outputs     = outputs
        self.parameters  = parameters

        # Open the database and load the objects to provide via the outputs
        self.db = bob.db.biowave_v1.Database()

        self.objs = sorted(self.db.objects(protocol=self.parameters['protocol'],
                                           groups=[self.parameters['group']],
                                           purposes=self.parameters.get('purpose', None),
                                           annotated_images=False,
                                           imagedir=self.root_folder),
                           key=lambda x: (x.client_id, x.id))

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        obj = self.objs[self.next_index - self.start_index]

        # Output: client_id (only provide data when the client_id change)
        if self.outputs['client_id'].isConnected() and \
           self.outputs['client_id'].last_written_data_index < self.next_index:

            client_end_index = get_client_end_index(self.objs, obj.client_id,
                                                    self.next_index,
                                                    self.start_index,
                                                    self.end_index)

            self.outputs['client_id'].write(
                {
                    'value': np.uint64(obj.client_id)
                },
                client_end_index
            )

        # Output: image (provide data at each iteration)
        if self.outputs['image'].isConnected():
            self.outputs['image'].write(
                {
                    'value': obj.load(directory=self.root_folder)
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


class TemplateView:
    """Outputs:
        - image: "{{ system_user.username }}/array_2d_uint8/1"
        - client_id: "{{ system_user.username }}/uint64/1"
        - model_id: "{{ system_user.username }}/text/1"

    Several "image" are associated with a given "model_id".
    Several "model_id" are associated with a given "client_id".

    --------------- --------------- --------------- --------------- --------------- ---------------
    |    image    | |    image    | |    image    | |    image    | |    image    | |    image    |
    --------------- --------------- --------------- --------------- --------------- ---------------
    ----------------------------------------------- -----------------------------------------------
    |                   model_id                  | |                   model_id                  |
    ----------------------------------------------- -----------------------------------------------
    -----------------------------------------------------------------------------------------------
    |                                          client_id                                          |
    -----------------------------------------------------------------------------------------------

    Note: for this particular database, there is only one "image"
    per "model_id".
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = root_folder
        self.outputs     = outputs
        self.parameters  = parameters

        # Open the database and load the objects to provide via the outputs
        self.db = bob.db.biowave_v1.Database()

        self.objs = sorted(self.db.objects(protocol=self.parameters['protocol'],
                                           purposes=self.parameters['purpose'],
                                           groups=[self.parameters['group']],
                                           annotated_images=False,
                                           imagedir=self.root_folder),
                           key=lambda x: (x.client_id, x.model_id, x.id))

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        obj = self.objs[self.next_index - self.start_index]

        # Output: model_id (only provide data when the model_id change)
        if self.outputs['model_id'].isConnected() and \
           self.outputs['model_id'].last_written_data_index < self.next_index:

            model_end_index = get_model_end_index(self.objs, obj.model_id,
                                                  self.next_index,
                                                  self.start_index,
                                                  self.end_index)

            self.outputs['model_id'].write(
                {
                    'text': obj.model_id.encode('utf-8')
                },
                model_end_index
            )

        # Output: client_id (only provide data when the client_id change)
        if self.outputs['client_id'].isConnected() and \
           self.outputs['client_id'].last_written_data_index < self.next_index:

            # Please pay attention - we can use the object CLIENT ID as long as we
            # don't use both hands data of the same client in the same protocol.
            client_end_index = get_client_end_index(self.objs, obj.client_id,
                                                    self.next_index,
                                                    self.start_index,
                                                    self.end_index)

            self.outputs['client_id'].write(
                {
                    'value': np.uint64(obj.client_id)
                },
                client_end_index
            )

        # Output: image (provide data at each iteration)
        if self.outputs['image'].isConnected():
            self.outputs['image'].write(
                {
                    'value': obj.load(directory=self.root_folder)
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


class ViewAnnotations:
    """Outputs:
        - image: "{{ system_user.username }}/array_2d_uint8/1"
        - vein_annotations: "{{ system_user.username }}/array_2d_uint8/1"
        - ROI_annotations: "{{ system_user.username }}/array_2d_uint8/1"
        - alignment_annotations: "{{ system_user.username }}/array_1d_coordinates/1"
        - client_id: "{{ system_user.username }}/uint64/1"

    One "vein_annotations" are associated with a given "image".
    One "ROI_annotations" are associated with a given "image".
    One "alignment_annotations" are associated with a given "image".
    Several "image" are associated with a given "client_id".

    ------------------------- ------------------------- ------------------------- -------------------------
    |         image         | |         image         | |         image         | |         image         |
    ------------------------- ------------------------- ------------------------- -------------------------
    ------------------------- ------------------------- ------------------------- -------------------------
    |    vein_annotations   | |    vein_annotations   | |    vein_annotations   | |    vein_annotations   |
    ------------------------- ------------------------- ------------------------- -------------------------
    ------------------------- ------------------------- ------------------------- -------------------------
    |    ROI_annotations    | |    ROI_annotations    | |    ROI_annotations    | |    ROI_annotations    |
    ------------------------- ------------------------- ------------------------- -------------------------
    ------------------------- ------------------------- ------------------------- -------------------------
    | alignment_annotations | | alignment_annotations | | alignment_annotations | | alignment_annotations |
    ------------------------- ------------------------- ------------------------- -------------------------
    --------------------------------------------------- ---------------------------------------------------
    |                     client_id                   | |                     client_id                   |
    --------------------------------------------------- ---------------------------------------------------
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = root_folder
        self.outputs     = outputs
        self.parameters  = parameters

        # Open the database and load the objects to provide via the outputs
        self.db = bob.db.biowave_v1.Database()

        self.objs = sorted(self.db.objects(protocol=self.parameters['protocol'],
                                           groups=[self.parameters['group']],
                                           purposes=self.parameters.get('purpose', None),
                                           annotated_images=True,
                                           imagedir=self.root_folder),
                           key=lambda x: (x.client_id, x.id))

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        obj = self.objs[self.next_index - self.start_index]

        # Output: client_id (only provide data when the client_id change)
        if self.outputs['client_id'].isConnected() and \
           self.outputs['client_id'].last_written_data_index < self.next_index:

            client_end_index = get_client_end_index(self.objs, obj.client_id,
                                                    self.next_index,
                                                    self.start_index,
                                                    self.end_index)

            self.outputs['client_id'].write(
                {
                    'value': np.uint64(obj.client_id)
                },
                client_end_index
            )

        # Output: vein_annotations (provide data at each iteration)
        if self.outputs['vein_annotations'].isConnected():
            self.outputs['vein_annotations'].write(
                {
                    'value': obj.construct_vein_image(directory=self.root_folder)
                },
                self.next_index
            )

        # Output: ROI_annotations (provide data at each iteration)
        if self.outputs['ROI_annotations'].isConnected():
            self.outputs['ROI_annotations'].write(
                {
                    'value': obj.construct_ROI_image(directory=self.root_folder)
                },
                self.next_index
            )

        # Output: alignment_annotations (provide data at each iteration)
        if self.outputs['alignment_annotations'].isConnected():
            alignment_annotations = obj.alignment_annotations(directory=self.root_folder)

            coordinate_list = []
            for annotation in alignment_annotations:
                coordinate = {
                    "y": np.int32(annotation[0]),
                    "x": np.int32(annotation[1])
                }
                coordinate_list.append(coordinate)

            self.outputs['alignment_annotations'].write(
                {
                    'value': coordinate_list
                },
                self.next_index
            )

        # Output: image (provide data at each iteration)
        if self.outputs['image'].isConnected():
            self.outputs['image'].write(
                {
                    'value': obj.load(directory=self.root_folder)
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


class TemplateViewAnnotations:
    """Outputs:
        - image: "{{ system_user.username }}/array_2d_uint8/1"
        - vein_annotations: "{{ system_user.username }}/array_2d_uint8/1"
        - ROI_annotations: "{{ system_user.username }}/array_2d_uint8/1"
        - alignment_annotations: "{{ system_user.username }}/array_1d_coordinates/1"
        - model_id: "{{ system_user.username }}/model_id/1"
        - client_id: "{{ system_user.username }}/uint64/1"

    One "vein_annotations" are associated with a given "image".
    One "ROI_annotations" are associated with a given "image".
    One "alignment_annotations" are associated with a given "image".
    Several "image" are associated with a given "model_id".
    Several "model_id" are associated with a given "client_id".

    ------------------------- ------------------------- ------------------------- -------------------------
    |         image         | |         image         | |         image         | |         image         |
    ------------------------- ------------------------- ------------------------- -------------------------
    ------------------------- ------------------------- ------------------------- -------------------------
    |    vein_annotations   | |    vein_annotations   | |    vein_annotations   | |    vein_annotations   |
    ------------------------- ------------------------- ------------------------- -------------------------
    ------------------------- ------------------------- ------------------------- -------------------------
    |    ROI_annotations    | |    ROI_annotations    | |    ROI_annotations    | |    ROI_annotations    |
    ------------------------- ------------------------- ------------------------- -------------------------
    ------------------------- ------------------------- ------------------------- -------------------------
    | alignment_annotations | | alignment_annotations | | alignment_annotations | | alignment_annotations |
    ------------------------- ------------------------- ------------------------- -------------------------
    --------------------------------------------------- ---------------------------------------------------
    |                     model_id                    | |                     model_id                    |
    --------------------------------------------------- ---------------------------------------------------
    -------------------------------------------------------------------------------------------------------
    |                                              client_id                                              |
    -------------------------------------------------------------------------------------------------------

    Note: for this particular database, there is only one "image"
    per "model_id".
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = root_folder
        self.outputs     = outputs
        self.parameters  = parameters

        # Open the database and load the objects to provide via the outputs
        self.db = bob.db.biowave_v1.Database()

        self.objs = sorted(self.db.objects(protocol=self.parameters['protocol'],
                                           purposes=self.parameters['purpose'],
                                           groups=[self.parameters['group']],
                                           annotated_images=True,
                                           imagedir=self.root_folder),
                           key=lambda x: (x.client_id, x.model_id, x.id))

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        obj = self.objs[self.next_index - self.start_index]

        # Output: model_id (only provide data when the model_id change)
        if self.outputs['model_id'].isConnected() and \
           self.outputs['model_id'].last_written_data_index < self.next_index:

            model_end_index = get_model_end_index(self.objs, obj.model_id,
                                                  self.next_index,
                                                  self.start_index,
                                                  self.end_index)

            self.outputs['model_id'].write(
                {
                    'text': obj.model_id.encode('utf-8')
                },
                model_end_index
            )

        # Output: client_id (only provide data when the client_id change)
        if self.outputs['client_id'].isConnected() and \
           self.outputs['client_id'].last_written_data_index < self.next_index:

            # Please pay attention - we can use the object CLIENT ID as long as we
            # don't use both hands data of the same client in the same protocol.
            client_end_index = get_client_end_index(self.objs, obj.client_id,
                                                    self.next_index,
                                                    self.start_index,
                                                    self.end_index)

            self.outputs['client_id'].write(
                {
                    'value': np.uint64(obj.client_id)
                },
                client_end_index
            )

        # Output: vein_annotations (provide data at each iteration)
        if self.outputs['vein_annotations'].isConnected():
            self.outputs['vein_annotations'].write(
                {
                    'value': obj.construct_vein_image(directory=self.root_folder)
                },
                self.next_index
            )

        # Output: ROI_annotations (provide data at each iteration)
        if self.outputs['ROI_annotations'].isConnected():
            self.outputs['ROI_annotations'].write(
                {
                    'value': obj.construct_ROI_image(directory=self.root_folder)
                },
                self.next_index
            )

        # Output: alignment_annotations (provide data at each iteration)
        if self.outputs['alignment_annotations'].isConnected():
            alignment_annotations = obj.alignment_annotations(directory=self.root_folder)

            coordinate_list = []
            for annotation in alignment_annotations:
                coordinate = {
                    "y": np.int32(annotation[0]),
                    "x": np.int32(annotation[1])
                }
                coordinate_list.append(coordinate)

            self.outputs['alignment_annotations'].write(
                {
                    'value': coordinate_list
                },
                self.next_index
            )

        # Output: image (provide data at each iteration)
        if self.outputs['image'].isConnected():
            self.outputs['image'].write(
                {
                    'value': obj.load(directory=self.root_folder)
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


class AnnotationBenchmark:
    """Outputs:
        - image: "{{ system_user.username }}/array_2d_uint8/1"
        - vein_annotations: "{{ system_user.username }}/array_2d_uint8/1"
        - ROI_annotations: "{{ system_user.username }}/array_2d_uint8/1"
        - alignment_annotations: "{{ system_user.username }}/array_1d_coordinates/1"

    One "vein_annotations" are associated with a given "image".
    One "ROI_annotations" are associated with a given "image".
    One "alignment_annotations" are associated with a given "image".

    ------------------------- ------------------------- ------------------------- -------------------------
    |         image         | |         image         | |         image         | |         image         |
    ------------------------- ------------------------- ------------------------- -------------------------
    ------------------------- ------------------------- ------------------------- -------------------------
    |    vein_annotations   | |    vein_annotations   | |    vein_annotations   | |    vein_annotations   |
    ------------------------- ------------------------- ------------------------- -------------------------
    ------------------------- ------------------------- ------------------------- -------------------------
    |    ROI_annotations    | |    ROI_annotations    | |    ROI_annotations    | |    ROI_annotations    |
    ------------------------- ------------------------- ------------------------- -------------------------
    ------------------------- ------------------------- ------------------------- -------------------------
    | alignment_annotations | | alignment_annotations | | alignment_annotations | | alignment_annotations |
    ------------------------- ------------------------- ------------------------- -------------------------
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = root_folder
        self.outputs     = outputs
        self.parameters  = parameters

        # Open the database and load the objects to provide via the outputs
        self.db = bob.db.biowave_v1.Database()

        self.objs = sorted(self.db.objects(protocol=self.parameters['protocol'],
                                           annotated_images=True,
                                           imagedir=self.root_folder),
                           key=lambda x: x.id)

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        obj = self.objs[self.next_index - self.start_index]

        # Output: vein_annotations (provide data at each iteration)
        if self.outputs['vein_annotations'].isConnected():
            self.outputs['vein_annotations'].write(
                {
                    'value': obj.construct_vein_image(directory=self.root_folder)
                },
                self.next_index
            )

        # Output: ROI_annotations (provide data at each iteration)
        if self.outputs['ROI_annotations'].isConnected():
            self.outputs['ROI_annotations'].write(
                {
                    'value': obj.construct_ROI_image(directory=self.root_folder)
                },
                self.next_index
            )

        # Output: alignment_annotations (provide data at each iteration)
        if self.outputs['alignment_annotations'].isConnected():
            alignment_annotations = obj.alignment_annotations(directory=self.root_folder)

            coordinate_list = []
            for annotation in alignment_annotations:
                coordinate = {
                    "y": np.int32(annotation[0]),
                    "x": np.int32(annotation[1])
                }
                coordinate_list.append(coordinate)

            self.outputs['alignment_annotations'].write(
                {
                    'value': coordinate_list
                },
                self.next_index
            )

        # Output: image (provide data at each iteration)
        if self.outputs['image'].isConnected():
            self.outputs['image'].write(
                {
                    'value': obj.load(directory=self.root_folder)
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


def setup_tests():
    # Install a mock methods
    def mock_load(obj, directory=''):
        return np.ndarray((10, 20), dtype=np.uint8)

    def mock_construct_vein_image(obj, directory=''):
        return np.ndarray((10, 20), dtype=np.uint8)

    def mock_construct_ROI_image(obj, directory=''):
        return np.ndarray((10, 20), dtype=np.uint8)

    bob.db.biowave_v1.models.File.load = mock_load
    bob.db.biowave_v1.models.File.construct_vein_image = mock_construct_vein_image
    bob.db.biowave_v1.models.File.construct_ROI_image = mock_construct_ROI_image


#----------------------------------------------------------


# Test the behavior of the views (on fake data)
if __name__ == '__main__':

    setup_tests()

    from beat.backend.python.database import DatabaseTester

    DatabaseTester('View', View,
        [
            'client_id',
            'image',
        ],
        parameters=dict(
            protocol="Idiap_1_1_R_BEAT_test",
            group="world",
        ),
    )

    DatabaseTester('TemplateView', TemplateView,
        [
            'client_id',
            'model_id',
            'image',
        ],
        parameters=dict(
            protocol="Idiap_1_1_R_BEAT_test",
            group="dev",
            purpose="enroll",
        ),
        irregular_outputs=[
            'model_id',
        ]
    )
