###############################################################################
#                                                                             #
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.examples module of the BEAT platform.         #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import os
import numpy as np
import bob.io.base
import bob.io.image
import bob.db.frgc


#----------------------------------------------------------


def get_client_end_index(objs, client_id, client_start_index,
                         start_index, end_index):
    client_end_index = client_start_index

    while client_end_index + 1 <= end_index:
        obj = objs[client_end_index + 1 - start_index]

        if isinstance(obj, tuple):
            obj = obj[1]

        if obj.client_id != client_id:
            return client_end_index

        client_end_index += 1

    return end_index


#----------------------------------------------------------


def get_value_end_index(objs, value, index_in_tuple, value_start_index,
                        start_index, end_index):
    value_end_index = value_start_index

    while value_end_index + 1 <= end_index:
        id = objs[value_end_index + 1 - start_index][index_in_tuple]

        if id != value:
            return value_end_index

        value_end_index += 1

    return end_index


#----------------------------------------------------------


class Train:
    """Outputs:
        - image: "{{ system_user.username }}/array_3d_uint8/1"
        - file_id: "{{ system_user.username }}/text/1"
        - eye_centers: "{{ system_user.username }}/eye_positions/1"
        - client_id: "{{ system_user.username }}/text/1"

    One "file_id" is associated with a given "image".
    One "eye_centers" is associated with a given "image".
    Several "image" are associated with a given "client_id".

    --------------- --------------- --------------- --------------- --------------- ---------------
    |    image    | |    image    | |    image    | |    image    | |    image    | |    image    |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    | eye_centers | | eye_centers | | eye_centers | | eye_centers | | eye_centers | | eye_centers |
    --------------- --------------- --------------- --------------- --------------- ---------------
    ----------------------------------------------- -----------------------------------------------
    |                   client_id                 | |                   client_id                 |
    ----------------------------------------------- -----------------------------------------------
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = root_folder
        self.outputs     = outputs

        # Open the database and load the objects to provide via the outputs
        self.db = bob.db.frgc.Database(original_directory=root_folder)

        self.objs = sorted(self.db.objects(protocol=parameters['protocol'],
                                           groups='world',
                                           mask_type=parameters['mask']),
                           key=lambda x: (x.client_id, x.id))

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        obj = self.objs[self.next_index - self.start_index]

        # Output: client_id (only provide data when the client_id change)
        if self.outputs['client_id'].isConnected() and \
           self.outputs['client_id'].last_written_data_index < self.next_index:

            client_end_index = get_client_end_index(self.objs, obj.client_id,
                                                    self.next_index,
                                                    self.start_index,
                                                    self.end_index)

            self.outputs['client_id'].write(
                {
                    'text': obj.client_id
                },
                client_end_index
            )

        # Output: file_id (provide data at each iteration)
        if self.outputs['file_id'].isConnected():
            self.outputs['file_id'].write(
                {
                    'text': obj.id
                },
                self.next_index
            )

        # Output: image (provide data at each iteration)
        if self.outputs['image'].isConnected():
            filename = obj.make_path(self.root_folder, '.jpg')
            if not os.path.exists(filename):
                path = obj.make_path(self.root_folder, '.JPG')

            self.outputs['image'].write(
                {
                    'value': bob.io.base.load(filename)
                },
                self.next_index
            )

        # Output: eye_centers (provide data at each iteration)
        if self.outputs['eye_centers'].isConnected():
            annotation = self.db.annotations(obj)

            self.outputs['eye_centers'].write({
                    'left': {
                        'y': np.int32(annotation['leye'][0]),
                        'x': np.int32(annotation['leye'][1]),
                    },
                    'right': {
                        'y': np.int32(annotation['reye'][0]),
                        'x': np.int32(annotation['reye'][1]),
                    }
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


class Templates:
    """Outputs:
        - image: "{{ system_user.username }}/array_3d_uint8/1"
        - file_id: "{{ system_user.username }}/text/1"
        - eye_centers: "{{ system_user.username }}/eye_positions/1"
        - template_id: "{{ system_user.username }}/uint64/1"
        - client_id: "{{ system_user.username }}/text/1"

    One "file_id" is associated with a given "image".
    One "eye_centers" is associated with a given "image".
    Several "image" are associated with a given "template_id".
    Several "template_id" are associated with a given "client_id".

    --------------- --------------- --------------- --------------- --------------- ---------------
    |    image    | |    image    | |    image    | |    image    | |    image    | |    image    |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    | eye_centers | | eye_centers | | eye_centers | | eye_centers | | eye_centers | | eye_centers |
    --------------- --------------- --------------- --------------- --------------- ---------------
    ----------------------------------------------- -----------------------------------------------
    |                  template_id                | |                  template_id                |
    ----------------------------------------------- -----------------------------------------------
    -----------------------------------------------------------------------------------------------
    |                                          client_id                                          |
    -----------------------------------------------------------------------------------------------
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = root_folder
        self.outputs     = outputs
        self.parameters  = parameters

        # Open the database and load the objects to provide via the outputs
        # Note: using the methods provided by the database is too slow, here is an
        # optimized version that does what we need
        self.db = bob.db.frgc.Database(original_directory=root_folder)

        model_files = bob.db.frgc.models.get_list(root_folder, 'dev', parameters['protocol'], 'enroll')
        mask = bob.db.frgc.models.get_mask(root_folder, parameters['protocol'], parameters['mask'])

        self.objs = []
        for model_index, model in enumerate(model_files):
            if (mask[:, model_index] > 0).any():
                for presentation in model.m_files:
                    self.objs.append( (model.m_model, bob.db.frgc.models.File(model.m_signature,
                                                                              presentation,
                                                                              model.m_files[presentation]) )
                    )

        self.objs = sorted(self.objs, key=lambda x: (x[0], x[1].client_id, x[1].id))

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        (template_id, obj) = self.objs[self.next_index - self.start_index]

        # Output: template_id (only provide data when the template_id change)
        if self.outputs['template_id'].isConnected() and \
           self.outputs['template_id'].last_written_data_index < self.next_index:

            template_end_index = get_value_end_index(self.objs, template_id, 0,
                                                     self.next_index,
                                                     self.start_index,
                                                     self.end_index)

            self.outputs['template_id'].write(
                {
                    'value': np.uint64(template_id)
                },
                template_end_index
            )

        # Output: client_id (only provide data when the client_id change)
        if self.outputs['client_id'].isConnected() and \
           self.outputs['client_id'].last_written_data_index < self.next_index:

            client_end_index = get_client_end_index(self.objs, obj.client_id,
                                                    self.next_index,
                                                    self.start_index,
                                                    self.end_index)

            self.outputs['client_id'].write(
                {
                    'text': obj.client_id
                },
                client_end_index
            )

        # Output: file_id (provide data at each iteration)
        if self.outputs['file_id'].isConnected():
            self.outputs['file_id'].write(
                {
                    'text': obj.id
                },
                self.next_index
            )

        # Output: image (provide data at each iteration)
        if self.outputs['image'].isConnected():
            filename = obj.make_path(self.root_folder, '.jpg')
            if not os.path.exists(filename):
                path = obj.make_path(self.root_folder, '.JPG')

            self.outputs['image'].write(
                {
                    'value': bob.io.base.load(filename)
                },
                self.next_index
            )

        # Output: eye_centers (provide data at each iteration)
        if self.outputs['eye_centers'].isConnected():
            annotation = self.db.annotations(obj)

            self.outputs['eye_centers'].write({
                    'left': {
                        'y': np.int32(annotation['leye'][0]),
                        'x': np.int32(annotation['leye'][1]),
                    },
                    'right': {
                        'y': np.int32(annotation['reye'][0]),
                        'x': np.int32(annotation['reye'][1]),
                    }
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


class Probes:
    """Outputs:
        - image: "{{ system_user.username }}/array_3d_uint8/1"
        - file_id: "{{ system_user.username }}/text/1"
        - eye_centers: "{{ system_user.username }}/eye_positions/1"
        - probe_id: "{{ system_user.username }}/text/1"
        - client_id: "{{ system_user.username }}/text/1"
        - template_ids: "{{ system_user.username }}/array_1d_uint64/1"

    One "file_id" is associated with a given "image".
    One "eye_centers" is associated with a given "image".
    One "probe_id" is associated with a given "image".
    Several "image" are associated with a given "client_id".
    Several "client_id" are associated with a given "template_ids".

    --------------- --------------- --------------- --------------- --------------- ---------------
    |    image    | |    image    | |    image    | |    image    | |    image    | |    image    |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    | eye_centers | | eye_centers | | eye_centers | | eye_centers | | eye_centers | | eye_centers |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    |   probe_id  | |   probe_id  | |   probe_id  | |   probe_id  | |   probe_id  | |   probe_id  |
    --------------- --------------- --------------- --------------- --------------- ---------------
    ----------------------------------------------- -----------------------------------------------
    |                   client_id                 | |                   client_id                 |
    ----------------------------------------------- -----------------------------------------------
    -----------------------------------------------------------------------------------------------
    |                                        template_ids                                         |
    -----------------------------------------------------------------------------------------------
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = root_folder
        self.outputs     = outputs
        self.parameters  = parameters

        # Open the database and load the objects to provide via the outputs
        # Note: using the methods provided by the database is too slow, here is an
        # optimized version that does what we need
        self.db = bob.db.frgc.Database(original_directory=root_folder)

        probe_files = bob.db.frgc.models.get_list(root_folder, 'dev', parameters['protocol'], 'probe')
        model_files = bob.db.frgc.models.get_list(root_folder, 'dev', parameters['protocol'], 'enroll')
        mask = bob.db.frgc.models.get_mask(root_folder, parameters['protocol'], parameters['mask'])

        template_ids = np.array([ x.m_model for x in model_files ])

        self.objs = []
        for probe_index, probe in enumerate(probe_files):
            template_indices = mask[probe_index, :].nonzero()[0]
            templates = sorted(template_ids.take(template_indices))

            for presentation in probe.m_files:
                self.objs.append( (templates, bob.db.frgc.models.File(probe.m_signature,
                                                                      presentation,
                                                                      probe.m_files[presentation]) )
                )

        self.objs = sorted(self.objs, key=lambda x: (len(x[0]), x[0], x[1].client_id, x[1].id))

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        (template_ids, obj) = self.objs[self.next_index - self.start_index]

        # Output: template_ids (only provide data when the template_ids change)
        if self.outputs['template_ids'].isConnected() and \
           self.outputs['template_ids'].last_written_data_index < self.next_index:

            template_ids_end_index = get_value_end_index(self.objs, template_ids, 0,
                                                         self.next_index,
                                                         self.start_index,
                                                         self.end_index)

            self.outputs['template_ids'].write(
                {
                    'value': np.uint64(template_ids)
                },
                template_ids_end_index
            )

        # Output: client_id (only provide data when the client_id change)
        if self.outputs['client_id'].isConnected() and \
           self.outputs['client_id'].last_written_data_index < self.next_index:

            client_end_index = get_client_end_index(self.objs, obj.client_id,
                                                    self.next_index,
                                                    self.start_index,
                                                    self.end_index)

            self.outputs['client_id'].write(
                {
                    'text': obj.client_id
                },
                client_end_index
            )

        # Output: probe_id (provide data at each iteration)
        if self.outputs['probe_id'].isConnected():
            self.outputs['probe_id'].write(
                {
                    'text': obj.id
                },
                self.next_index
            )

        # Output: file_id (provide data at each iteration)
        if self.outputs['file_id'].isConnected():
            self.outputs['file_id'].write(
                {
                    'text': obj.id
                },
                self.next_index
            )

        # Output: image (provide data at each iteration)
        if self.outputs['image'].isConnected():
            filename = obj.make_path(self.root_folder, '.jpg')
            if not os.path.exists(filename):
                path = obj.make_path(self.root_folder, '.JPG')

            self.outputs['image'].write(
                {
                    'value': bob.io.base.load(filename)
                },
                self.next_index
            )

        # Output: eye_centers (provide data at each iteration)
        if self.outputs['eye_centers'].isConnected():
            annotation = self.db.annotations(obj)

            self.outputs['eye_centers'].write({
                    'left': {
                        'y': np.int32(annotation['leye'][0]),
                        'x': np.int32(annotation['leye'][1]),
                    },
                    'right': {
                        'y': np.int32(annotation['reye'][0]),
                        'x': np.int32(annotation['reye'][1]),
                    }
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


def setup_tests():
    # Install a mock load function for the images
    def mock_load(root_folder):
        return np.ndarray((3, 10, 20), dtype=np.uint8)

    # Install a mock os.path.exists function
    def mock_exists(path):
        return True

    bob.io.base.load = mock_load
    os.path.exists = mock_exists


#----------------------------------------------------------


# Test the behavior of the views (on fake data)
if __name__ == '__main__':

    setup_tests()

    # Note: This database can't be tested without the actual data, since
    # some files are needed by this implementation

    from beat.backend.python.database import DatabaseTester

    DatabaseTester('Train', Train,
        [
            'client_id',
            'file_id',
            'eye_centers',
            'image',
        ],
        parameters=dict(
            protocol='2.0.1',
            mask='maskI',
        ),
    )

    DatabaseTester('Templates', Templates,
        [
            'client_id',
            'template_id',
            'file_id',
            'eye_centers',
            'image',
        ],
        parameters=dict(
            protocol='2.0.1',
            mask='maskI',
        ),
        irregular_outputs=[
            'client_id',
        ]
    )

    DatabaseTester('Probes', Probes,
        [
            'template_ids',
            'client_id',
            'probe_id',
            'file_id',
            'eye_centers',
            'image',
        ],
        parameters=dict(
            protocol='2.0.1',
            mask='maskI',
        ),
        irregular_outputs=[
            'template_ids',
            'client_id',
        ]
    )
