import string
import numpy
import bob.db.kboc16


keystroke_feature = ['', 'timestamps', 'key_events']
#keystroke_type = ['', 'given_name', 'family_name', 'email', 'nationality', 'id_number']

def keystroke_reader(filename):
  data = {}
  valuesTime = []
  valuesKey = []
  for line in open(filename, 'r').readlines():
    parts = string.split(line)
    valuesTime.append(numpy.int32(parts[1]))
    valuesKey.append(parts[0])

  data[keystroke_feature[1]] = valuesTime
  data[keystroke_feature[2]] = valuesKey

  return data


class Templates:
    # Reasoning: Each client may have a number of models in certain databases.
    # So, each model receives an unique identifier. Those identifiers are linked
    # to the client identifier and contain a number of keystrokes to generated
    # the model from.

    def setup(self, root_folder, outputs, parameters, force_start_index=None, force_end_index=None):
        self.root_folder            = root_folder
        self.outputs                = outputs
        self.parameters             = parameters

        self.db                     = bob.db.kboc16.Database()

        self.template_ids           = sorted([str(v) for v in self.db.model_ids(groups='eval',
                                                                protocol=parameters['protocol'])])
        self.objs                   = None

        self.current_template_index = 0
        self.current_obj_index      = 0
        self.next_index             = 0

        self.force_start_index = force_start_index
        self.force_end_index   = force_end_index

        # Retrieve only 'useful' data
        ### End index
        #if self.force_end_index is not None:
        #    self.objs = self.objs[:self.force_end_index+1]
        ### Start index
        if self.force_start_index is None:
            self.force_start_index = 0
        while self.next_index < self.force_start_index:
            self.next()

        return True


    def done(self):
        return (self.current_template_index >= len(self.template_ids)) or (self.force_end_index is not None and self.force_end_index < self.next_index)


    def next(self):
        if self.objs is None:
            # probe for the specific objects concerning a given client
            self.objs = sorted(self.db.objects(model_ids=[self.template_ids[self.current_template_index]],
              groups='eval',
              protocol=self.parameters['protocol'],
              purposes='enrol'), key=lambda x: x.id)

            if self.force_start_index <= self.next_index and (self.force_end_index is None or self.force_end_index >= self.next_index):
                # For this database, 'self.template_ids[self.current_template_index]' corresponds to 'obj.client_id'
                if self.outputs['template_id'].isConnected():
                    self.outputs['template_id'].write({'text': str(self.template_ids[self.current_template_index])}, self.next_index+len(self.objs)-1)

                if self.outputs['client_id'].isConnected():
                    self.outputs['client_id'].write({'text': str(self.objs[0].client_id)}, self.next_index+len(self.objs)-1)

        obj = self.objs[self.current_obj_index]

        if self.outputs['file_id'].isConnected() or self.outputs['keystroke'].isConnected():
            if self.force_start_index <= self.next_index and (self.force_end_index is None or self.force_end_index >= self.next_index):
                if self.outputs['file_id'].isConnected():
                    self.outputs['file_id'].write({'value': numpy.uint64(obj.id)}, self.next_index)

                if self.outputs['keystroke'].isConnected():
                    data = keystroke_reader(obj.make_path(self.root_folder, '.txt'))
                    self.outputs['keystroke'].write(data, self.next_index)

            self.next_index += 1
            self.current_obj_index += 1

        else:
            self.next_index += len(self.objs)
            self.current_obj_index = len(self.objs)

        if self.current_obj_index == len(self.objs):
            self.objs = None
            self.current_obj_index = 0
            self.current_template_index += 1

        return True


class Probes:
    # Reasoning: Each client may have a number of probes. Each probe may be
    # composed of any number of keystrokes. So, each probe receives an unique
    # identifier. Those identifiers are linked to the client identifier and
    # contain a number of keystrokes to generated the match/comparison score.
    # Each probe must be matched against a number of models defined by a list
    # of client identifiers.

    def setup(self, root_folder, outputs, parameters, force_start_index=None, force_end_index=None):
        self.root_folder = root_folder
        self.outputs     = outputs
        self.parameters  = parameters

        self.db          = bob.db.kboc16.Database()

        self.template_ids = sorted([str(v) for v in self.db.model_ids(groups='eval',
                                                                      protocol=parameters['protocol'])])
        self.objs         = sorted(self.db.objects(groups='eval', protocol=self.parameters['protocol'], purposes='probe'), key=lambda x: x.id)

        self.template_probes        = {}
        for t_id in self.template_ids:
            self.template_probes[t_id] = [p.id for p in sorted(self.db.objects(model_ids=[t_id], groups='eval', protocol=self.parameters['protocol'], purposes='probe'), key=lambda x: x.id)]

        self.next_index  = 0

        self.force_start_index = force_start_index
        self.force_end_index   = force_end_index

        # Retrieve only 'useful' data
        ### End index
        if self.force_end_index is not None:
            self.objs = self.objs[:self.force_end_index+1]
        ### Start index
        if self.force_start_index is not None:
            self.objs = self.objs[self.force_start_index:]
            self.next_index = self.force_start_index
        else:
            self.force_start_index = 0

        return True


    def done(self):
        return (self.next_index-self.force_start_index >= len(self.objs))


    def next(self):

        obj = self.objs[self.next_index-self.force_start_index]

        if self.outputs['file_id'].isConnected():
            self.outputs['file_id'].write({'value': numpy.uint64(obj.id)}, self.next_index)

        if self.outputs['probe_id'].isConnected():
            self.outputs['probe_id'].write({'value': numpy.uint64(obj.id)}, self.next_index)

        if self.outputs['client_id'].isConnected():
            self.outputs['client_id'].write({'text': str(obj.client_id)}, self.next_index)

        if self.outputs['template_ids'].isConnected():
            data = {
              'text': [t_id for t_id in self.template_ids if obj.id in self.template_probes[t_id]]
            }
            self.outputs['template_ids'].write(data, self.next_index)

        if self.outputs['keystroke'].isConnected():
            data = keystroke_reader(obj.make_path(self.root_folder, '.txt'))
            self.outputs['keystroke'].write(data, self.next_index)

        self.next_index += 1

        return True
