###############################################################################
#                                                                             #
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.examples module of the BEAT platform.         #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import numpy as np
import bob.db.mnist


#----------------------------------------------------------


def get_label_end_index(objs, label, label_start_index,
                        start_index, end_index):
    label_end_index = label_start_index

    while label_end_index + 1 <= end_index:
        label_ = objs[label_end_index + 1 - start_index][1]

        if label_ != label:
            return label_end_index

        label_end_index += 1

    return end_index


#----------------------------------------------------------


class View:
    """Outputs:
        - image: "{{ system_user.username }}/array_2d_uint8/1"
        - id: "{{ system_user.username }}/uint64/1"
        - class_id: "{{ system_user.username }}/uint64/1"

    One "id" is associated with a given "image".
    Several "image" are associated with a given "class_id".

    --------------- --------------- --------------- --------------- --------------- ---------------
    |    image    | |    image    | |    image    | |    image    | |    image    | |    image    |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    |     id      | |     id      | |     id      | |     id      | |     id      | |     id      |
    --------------- --------------- --------------- --------------- --------------- ---------------
    ----------------------------------------------- -----------------------------------------------
    |                   class_id                  | |                   class_id                  |
    ----------------------------------------------- -----------------------------------------------
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = root_folder
        self.outputs     = outputs
        self.parameters  = parameters

        # Open the database and load the objects to provide via the outputs
        self.db = bob.db.mnist.Database(data_dir=self.root_folder)

        features, labels = self.db.data(groups=parameters['group'])

        self.objs = sorted([ (features[i], labels[i]) for i in range(len(features)) ],
                           key=lambda x: x[1])

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        features, label = self.objs[self.next_index - self.start_index]

        # Output: class_id (only provide data when the class_id change)
        if self.outputs['class_id'].isConnected() and \
           self.outputs['class_id'].last_written_data_index < self.next_index:

            label_end_index = get_label_end_index(self.objs, label,
                                                  self.next_index,
                                                  self.start_index,
                                                  self.end_index)

            self.outputs['class_id'].write(
                {
                    'value': np.uint64(label)
                },
                label_end_index
            )

        # Output: id (provide data at each iteration)
        if self.outputs['id'].isConnected():
            self.outputs['id'].write(
                {
                    'value': np.uint64(self.next_index)
                },
                self.next_index
            )

        # Output: image (provide data at each iteration)
        if self.outputs['image'].isConnected():
            self.outputs['image'].write(
                {
                    'value': features.reshape((28, 28))
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


def setup_tests():
    pass


#----------------------------------------------------------


# Test the behavior of the views (on fake data)
if __name__ == '__main__':

    setup_tests()

    # Note: This database can't be tested without the actual data, since
    # the actual files are needed by this implementation

    from beat.backend.python.database import DatabaseTester

    DatabaseTester('View', View,
        [
            'class_id',
            'id',
            'image',
        ],
        parameters=dict(
            group='train',
        ),
    )
