###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.examples module of the BEAT platform.         #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import numpy
import xbob.db.nist_sre12
import xbob.sox


class Train:

    def setup(self, root_folder, outputs, parameters, force_start_index=None, force_end_index=None):
        self.root_folder = root_folder
        self.outputs     = outputs

        self.db          = xbob.db.nist_sre12.Database()
        self.objs        = sorted(self.db.objects(groups='world', protocol=parameters['protocol'], world_gender=parameters['world_gender']))
        self.next_index  = 0

        self.force_start_index = force_start_index
        self.force_end_index   = force_end_index

        # Retrieve only 'useful' data
        ### End index
        if self.force_end_index is not None:
            self.objs = self.objs[:self.force_end_index+1]
        ### Start index
        if self.force_start_index is not None:
            self.objs = self.objs[self.force_start_index:]
            self.next_index = self.force_start_index
        else:
            self.force_start_index = 0

        return True


    def done(self):
        return (self.next_index-self.force_start_index >= len(self.objs))


    def next(self):

        obj = self.objs[self.next_index-self.force_start_index]

        if self.outputs['file_id'].isConnected():
            self.outputs['file_id'].write({'text': obj.id}, self.next_index)

        if self.outputs['client_id'].isConnected():
            self.outputs['client_id'].write({'text': obj.client_id}, self.next_index)

        if self.outputs['speech'].isConnected():
            infilename = obj.make_path(self.root_folder,  '.sph')
            audio = xbob.sox.reader(infilename)
            (rate, data_) = audio.load()
            data = numpy.cast['float'](data_[0]*pow(2,15)) # pow(2,15) is used to get the same native format as for scipy.io.wavfile.read
            self.outputs['speech'].write({'value': numpy.uint64(data)}, self.next_index)

        self.next_index += 1

        return True


class Templates:
    # Reasoning: Each client may have a number of models in certain databases.
    # So, each model receives an unique identifier. Those identifiers are linked
    # to the client identifier and contain a number of images to generated the
    # model from.

    def setup(self, root_folder, outputs, parameters, force_start_index=None, force_end_index=None):
        self.root_folder            = root_folder
        self.outputs                = outputs

        self.db                     = xbob.db.nist_sre12.Database()
        self.template_ids           = sorted(self.db.model_ids(groups='dev', protocol=parameters['protocol']))
        self.objs                   = None

        self.current_template_index = 0
        self.current_obj_index      = 0
        self.next_index             = 0

        self.force_start_index = force_start_index
        self.force_end_index   = force_end_index

        # Retrieve only 'useful' data
        ### End index
        #if self.force_end_index is not None:
        #    self.objs = self.objs[:self.force_end_index+1]
        ### Start index
        if self.force_start_index is None:
            self.force_start_index = 0
        while self.next_index < self.force_start_index:
            self.next()

        return True


    def done(self):
        return (self.current_template_index >= len(self.template_ids)) or (self.force_end_index is not None and self.force_end_index < self.next_index)


    def next(self):
        if self.objs is None:
            # probe for the specific objects concerning a given client
            self.objs = sorted(self.db.objects(model_ids=[self.template_ids[self.current_template_index]],
                                               groups='dev', purposes='enrol'), key=lambda x: x.id)

            if self.force_start_index <= self.next_index and (self.force_end_index is None or self.force_end_index >= self.next_index):
                # For this database, 'self.template_ids[self.current_template_index]' corresponds to 'obj.client_id'
                if self.outputs['template_id'].isConnected():
                    self.outputs['template_id'].write({'text': self.template_ids[self.current_template_index]}, self.next_index+len(self.objs)-1)

                if self.outputs['client_id'].isConnected():
                    self.outputs['client_id'].write({'text': self.objs[0].client_id}, self.next_index+len(self.objs)-1)

        obj = self.objs[self.current_obj_index]

        if self.outputs['file_id'].isConnected() or self.outputs['speech'].isConnected():
            if self.force_start_index <= self.next_index and (self.force_end_index is None or self.force_end_index >= self.next_index):
                if self.outputs['file_id'].isConnected():
                    self.outputs['file_id'].write({'text': obj.id}, self.next_index)

                if self.outputs['speech'].isConnected():
                    infilename = obj.make_path(self.root_folder,  '.sph')
                    audio = xbob.sox.reader(infilename)
                    (rate, data_) = audio.load()
                    data = numpy.cast['float'](data_[0]*pow(2,15)) # pow(2,15) is used to get the same native format as for scipy.io.wavfile.read
                    self.outputs['speech'].write({'value': numpy.uint64(data)}, self.next_index)

            self.next_index += 1
            self.current_obj_index += 1
        else:
            self.next_index += len(self.objs)
            self.current_obj_index = len(self.objs)

        if self.current_obj_index == len(self.objs):
            self.objs = None
            self.current_obj_index = 0
            self.current_template_index += 1

        return True


class Probes:
    # Reasoning: Each client may have a number of probes. Each probe may be
    # composed of any number of images. So, each probe receives an unique
    # identifier. Those identifiers are linked to the client identifier and
    # contain a number of images to generated the match/comparison score. Each
    # probe must be matched against a number of models defined by a list of
    # client identifiers.

    def setup(self, root_folder, outputs, parameters, force_start_index=None, force_end_index=None):
        self.root_folder = root_folder
        self.outputs     = outputs

        self.db          = xbob.db.nist_sre12.Database()
        self.objs        = sorted(self.db.objects(groups='dev', purposes='probe', protocol=parameters['protocol']), key=lambda x: x.id)
        self.next_index  = 0

        self.force_start_index = force_start_index
        self.force_end_index   = force_end_index

        # Retrieve only 'useful' data
        ### End index
        if self.force_end_index is not None:
            self.objs = self.objs[:self.force_end_index+1]
        ### Start index
        if self.force_start_index is not None:
            self.objs = self.objs[self.force_start_index:]
            self.next_index = self.force_start_index
        else:
            self.force_start_index = 0

        return True


    def done(self):
        return (self.next_index-self.force_start_index >= len(self.objs))


    def next(self):

        obj = self.objs[self.next_index-self.force_start_index]

        if self.outputs['file_id'].isConnected():
            self.outputs['file_id'].write({'text': obj.id}, self.next_index)

        if self.outputs['probe_id'].isConnected():
            self.outputs['probe_id'].write({'text': obj.id}, self.next_index)

        if self.outputs['client_id'].isConnected():
            self.outputs['client_id'].write({'text': obj.client_id}, self.next_index)

        if self.outputs['template_ids'].isConnected():
              data = {
                  'text': numpy.array(self.db.model_ids(groups='dev'), dtype=numpy.string_)
              }
              self.outputs['template_ids'].write(data, self.next_index)

        if self.outputs['speech'].isConnected():
            infilename = obj.make_path(self.root_folder,  '.sph')
            audio = xbob.sox.reader(infilename)
            (rate, data_) = audio.load()
            data = numpy.cast['float'](data_[0]*pow(2,15)) # pow(2,15) is used to get the same native format as for scipy.io.wavfile.read
            self.outputs['speech'].write({'value': numpy.uint64(data)}, self.next_index)

        self.next_index += 1

        return True
