###############################################################################
#                                                                             #
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.examples module of the BEAT platform.         #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import numpy as np
import bob.db.putvein
import bob.ip.color
import os


#----------------------------------------------------------


def get_client_end_index(objs, client_id, client_start_index,
                         start_index, end_index):
    client_end_index = client_start_index

    while client_end_index + 1 <= end_index:
        obj = objs[client_end_index + 1 - start_index]

        if isinstance(obj, tuple):
            obj = obj[1]

        if obj.get_client_id() != client_id:
            return client_end_index

        client_end_index += 1

    return end_index


#----------------------------------------------------------


def get_model_end_index(objs, model_id, model_start_index,
                        start_index, end_index):
    model_end_index = model_start_index

    while model_end_index + 1 <= end_index:
        id = objs[model_end_index + 1 - start_index][0]

        if id != model_id:
            return model_end_index

        model_end_index += 1

    return end_index


#----------------------------------------------------------


class View:
    """Outputs:
        - image: "{{ system_user.username }}/array_2d_uint8/1"
        - client_id: "{{ system_user.username }}/uint64/1"

    Several "image" are associated with a given "client_id"

    --------- --------- --------- --------- --------- ---------
    | image | | image | | image | | image | | image | | image |
    --------- --------- --------- --------- --------- ---------
    ----------------------------- -----------------------------
    |          client_id        | |          client_id        |
    ----------------------------- -----------------------------
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = os.path.join(root_folder, '')
        self.outputs = outputs
        self.parameters = parameters

        # Open the database and load the objects to provide via the outputs
        self.db = bob.db.putvein.Database()

        self.objs = sorted(self.db.objects(protocol=self.parameters['protocol'],
                                           purposes=self.parameters.get('purpose', None),
                                           groups=[self.parameters['group']],
                                           kinds=[self.parameters['kind']]),
                           key=lambda x: x.client_id)

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        obj = self.objs[self.next_index - self.start_index]

        # Output: client_id (only provide data when the client_id change)
        if self.outputs['client_id'].isConnected() and \
           self.outputs['client_id'].last_written_data_index < self.next_index:

            # Please pay attention - we can use the object CLIENT ID as long as we
            # don't use both hands data of the same client in the same protocol.
            client_id = obj.get_client_id()

            client_end_index = get_client_end_index(self.objs, client_id,
                                                    self.next_index,
                                                    self.start_index,
                                                    self.end_index)

            self.outputs['client_id'].write(
                {
                    'value': np.uint64(client_id)
                },
                client_end_index
            )

        # Output: image (provide data at each iteration)
        if self.outputs['image'].isConnected():
            """
            The image returned by the ``bob.db.putvein`` is RGB (with shape
            (3, 768, 1024)). This method converts image to a grayscale
            (shape (768, 1024)) and then rotates image by 270 deg so that
            images can be used with ``bob.bio.vein`` algorithms designed for
            the ``bob.db.biowave_v1`` database.
            Output images dimensions: (1024, 768).
            """
            color_image = obj.load(self.root_folder)
            grayscale_image = bob.ip.color.rgb_to_gray(color_image)
            grayscale_image = np.rot90(grayscale_image, k=3)

            self.outputs['image'].write(
                {
                    'value': grayscale_image
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


class TemplateView:
    """Outputs:
        - image: "{{ system_user.username }}/array_2d_uint8/1"
        - model_id: "{{ system_user.username }}/text/1"
        - client_id: "{{ system_user.username }}/uint64/1"

    Several "image" are associated with a given "model_id".
    Several "model_id" are associated with a given "client_id".

    --------------- --------------- --------------- --------------- --------------- ---------------
    |    image    | |    image    | |    image    | |    image    | |    image    | |    image    |
    --------------- --------------- --------------- --------------- --------------- ---------------
    ----------------------------------------------- -----------------------------------------------
    |                   model_id                  | |                   model_id                  |
    ----------------------------------------------- -----------------------------------------------
    -----------------------------------------------------------------------------------------------
    |                                          client_id                                          |
    -----------------------------------------------------------------------------------------------

    Note: for this particular database, there is only one "image"
    per "model_id".
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = os.path.join(root_folder, '')
        self.outputs = outputs
        self.parameters = parameters

        # Open the database and load the objects to provide via the outputs
        self.db = bob.db.putvein.Database()

        model_ids = self.db.model_ids(protocol=self.parameters['protocol'],
                                      groups=[self.parameters['group']],
                                      kinds=[self.parameters['kind']])

        self.objs = []

        for model_id in model_ids:
            objs = self.db.objects(protocol=self.parameters['protocol'],
                                   purposes=self.parameters.get('purpose', None),
                                   groups=[self.parameters['group']],
                                   kinds=[self.parameters['kind']],
                                   model_ids=[model_id])

            self.objs.extend([ (model_id, obj) for obj in objs ])

        self.objs = sorted(self.objs, key=lambda x: (x[1].client_id, x[0], x[1].id))

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        (model_id, obj) = self.objs[self.next_index - self.start_index]

        # Output: model_id (only provide data when the model_id change)
        if self.outputs['model_id'].isConnected() and \
           self.outputs['model_id'].last_written_data_index < self.next_index:

            model_end_index = get_model_end_index(self.objs, model_id,
                                                  self.next_index,
                                                  self.start_index,
                                                  self.end_index)

            self.outputs['model_id'].write(
                {
                    'text': model_id.encode('utf-8')
                },
                model_end_index
            )

        # Output: client_id (only provide data when the client_id change)
        if self.outputs['client_id'].isConnected() and \
           self.outputs['client_id'].last_written_data_index < self.next_index:

            # Please pay attention - we can use the object CLIENT ID as long as we
            # don't use both hands data of the same client in the same protocol.
            client_end_index = get_client_end_index(self.objs, obj.client_id,
                                                    self.next_index,
                                                    self.start_index,
                                                    self.end_index)

            self.outputs['client_id'].write(
                {
                    'value': np.uint64(obj.client_id)
                },
                client_end_index
            )

        # Output: image (provide data at each iteration)
        if self.outputs['image'].isConnected():
            """
            The image returned by the ``bob.db.putvein`` is RGB (with shape
            (3, 768, 1024)). This method converts image to a grayscale
            (shape (768, 1024)) and then rotates image by 270 deg so that
            images can be used with ``bob.bio.vein`` algorithms designed for
            the ``bob.db.biowave_v1`` database.
            Output images dimensions: (1024, 768).
            """
            color_image = obj.load(self.root_folder)
            grayscale_image = bob.ip.color.rgb_to_gray(color_image)
            grayscale_image = np.rot90(grayscale_image, k=3)

            self.outputs['image'].write(
                {
                    'value': grayscale_image
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


def setup_tests():
    # Install a mock methods
    def mock_load(obj, directory):
        return np.ndarray((3, 10, 20), dtype=np.uint8)

    bob.db.putvein.models.File.load = mock_load


#----------------------------------------------------------


# Test the behavior of the views (on fake data)
if __name__ == '__main__':

    setup_tests()

    from beat.backend.python.database import DatabaseTester

    DatabaseTester('View', View,
        [
            'client_id',
            'image',
        ],
        dict(
            protocol = 'LR_4',
            kind = 'wrist',
            group = 'dev',
            purpose = 'probe',
        )
    )

    DatabaseTester('TemplateView', TemplateView,
        [
            'model_id',
            'client_id',
            'image',
        ],
        dict(
            protocol = 'LR_4',
            kind = 'wrist',
            group = 'dev',
            purpose = 'enroll',
        )
    )
