###############################################################################
#                                                                             #
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.examples module of the BEAT platform.         #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import numpy as np
import bob.io.base
import bob.io.video
import bob.db.replay


from bob.db.replay.driver import Interface
INFO = Interface()
SQLITE_FILE = INFO.files()[0]


#----------------------------------------------------------


def get_client_end_index(objs, client_id, client_start_index,
                         start_index, end_index):
    client_end_index = client_start_index

    while client_end_index + 1 <= end_index:
        obj = objs[client_end_index + 1 - start_index]

        if isinstance(obj, tuple):
            obj = obj[-1]

        if obj.client_id != client_id:
            return client_end_index

        client_end_index += 1

    return end_index


#----------------------------------------------------------


def get_value_end_index(objs, value, index_in_tuple, value_start_index,
                        start_index, end_index):
    value_end_index = value_start_index

    while value_end_index + 1 <= end_index:
        obj_value = objs[value_end_index + 1 - start_index][index_in_tuple]

        if obj_value != value:
            return value_end_index

        value_end_index += 1

    return end_index


#----------------------------------------------------------


class All:
    """Outputs:
        - video: "{{ system_user.username }}/array_4d_uint8/1"
        - file_id: "{{ system_user.username }}/uint64/1"
        - annotations: "{{ system_user.username }}/bounding_box_video/1"
        - client_id: "{{ system_user.username }}/uint64/1"
        - attack_support: "{{ system_user.username }}/text/1"
        - class: "{{ system_user.username }}/text/1"

    One "file_id" is associated with a given "video".
    One "annotations" is associated with a given "video".
    Several "video" are associated with a given "client_id".
    Several "client_id" are associated with a given "class".
    Several "attack_support" are associated with a given "class".

    --------------- --------------- --------------- --------------- --------------- ---------------
    |    video    | |    video    | |    video    | |    video    | |    video    | |    video    |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    | annotations | | annotations | | annotations | | annotations | | annotations | | annotations |
    --------------- --------------- --------------- --------------- --------------- ---------------
    ------------------------------- ------------------------------- -------------------------------
    |          client_id          | |          client_id          | |          client_id          |
    ------------------------------- ------------------------------- -------------------------------
    --------------------------------------------------------------- -------------------------------
    |                        attack_support                       | |        attack_support       |
    --------------------------------------------------------------- -------------------------------
    -----------------------------------------------------------------------------------------------
    |                                             class                                           |
    -----------------------------------------------------------------------------------------------
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = root_folder
        self.outputs     = outputs
        self.parameters  = parameters

        # Open the database and load the objects to provide via the outputs
        self.db = bob.db.replay.Database()

        self.objs = []

        if self.parameters['enroll']:
            self.objs.extend([ ('enroll', '', x)
                               for x in sorted(self.db.objects(protocol=parameters['protocol'],
                                                               groups=parameters['group'],
                                                               cls='enroll'),
                                               key=lambda x: (x.client_id, x.id))
                             ])

        self.objs.extend([ ('real', '', x)
                           for x in sorted(self.db.objects(protocol=parameters['protocol'],
                                                           groups=parameters['group'],
                                                           cls='real'),
                                           key=lambda x: (x.client_id, x.id))
                         ])

        self.objs.extend([ ('attack', 'fixed', x)
                           for x in sorted(self.db.objects(protocol=parameters['protocol'],
                                                           groups=parameters['group'],
                                                           cls='attack',
                                                           support='fixed'),
                                           key=lambda x: (x.client_id, x.id))
                         ])

        self.objs.extend([ ('attack', 'hand', x)
                           for x in sorted(self.db.objects(protocol=parameters['protocol'],
                                                           groups=parameters['group'],
                                                           cls='attack',
                                                           support='hand'),
                                           key=lambda x: (x.client_id, x.id))
                         ])

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        (cls, attack_support, obj) = self.objs[self.next_index - self.start_index]

        # Output: class (only provide data when the class change)
        if self.outputs['class'].isConnected() and \
           self.outputs['class'].last_written_data_index < self.next_index:

            end_index = get_value_end_index(self.objs, cls, 0,
                                            self.next_index,
                                            self.start_index,
                                            self.end_index)

            self.outputs['class'].write(
                {
                    'text': cls
                },
                end_index
            )

        # Output: attack_support (only provide data when the attack_support change)
        if self.outputs['attack_support'].isConnected() and \
           self.outputs['attack_support'].last_written_data_index < self.next_index:

            end_index = get_value_end_index(self.objs, attack_support, 1,
                                            self.next_index,
                                            self.start_index,
                                            self.end_index)

            self.outputs['attack_support'].write(
                {
                    'text': attack_support
                },
                end_index
            )

        # Output: client_id (only provide data when the client_id change)
        if self.outputs['client_id'].isConnected() and \
           self.outputs['client_id'].last_written_data_index < self.next_index:

            end_index = get_client_end_index(self.objs, obj.client_id,
                                             self.next_index,
                                             self.start_index,
                                             self.end_index)

            self.outputs['client_id'].write(
                {
                    'value': np.uint64(obj.client_id)
                },
                end_index
            )

        # Output: annotations (provide data at each iteration)
        if self.outputs['annotations'].isConnected():
            annotations = obj.bbx(self.root_folder)

            annotations_list = []
            for i in range(annotations.shape[0]):
                annotations_list.append({
                    'frame_id': np.uint64(annotations[i][0]),
                    'top-left-x': np.int32(annotations[i][1]),
                    'top-left-y': np.int32(annotations[i][2]),
                    'width': np.int32(annotations[i][3]),
                    'height': np.int32(annotations[i][4])
                })

            self.outputs['annotations'].write(
                {
                    'value': annotations_list
                },
                self.next_index
            )

        # Output: file_id (provide data at each iteration)
        if self.outputs['file_id'].isConnected():
            self.outputs['file_id'].write(
                {
                    'value': np.uint64(obj.id)
                },
                self.next_index
            )

        # Output: image (provide data at each iteration)
        if self.outputs['video'].isConnected():
            self.outputs['video'].write(
                {
                    'value': bob.io.base.load(obj.videofile(self.root_folder))
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


class Templates:
    """Outputs:
        - video: "{{ system_user.username }}/array_4d_uint8/1"
        - file_id: "{{ system_user.username }}/uint64/1"
        - annotations: "{{ system_user.username }}/bounding_box_video/1"
        - template_id: "{{ system_user.username }}/uint64/1"
        - client_id: "{{ system_user.username }}/uint64/1"

    One "file_id" is associated with a given "video".
    One "annotations" is associated with a given "video".
    Several "video" are associated with a given "template_id".
    Several "template_id" are associated with a given "client_id".

    --------------- --------------- --------------- ---------------
    |    video    | |    video    | |    video    | |    video    |
    --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- ---------------
    |   file_id   | |   file_id   | |   file_id   | |   file_id   |
    --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- ---------------
    | annotations | | annotations | | annotations | | annotations |
    --------------- --------------- --------------- ---------------
    ------------------------------- -------------------------------
    |         template_id         | |         template_id         |
    ------------------------------- -------------------------------
    ---------------------------------------------------------------
    |                          client_id                          |
    ---------------------------------------------------------------

    Note: for this particular database, there is only one "template_id"
    per "client_id".
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = root_folder
        self.outputs     = outputs
        self.parameters  = parameters

        # Open the database and load the objects to provide via the outputs
        self.db = bob.db.replay.Database()

        self.objs = sorted(self.db.objects(protocol=self.parameters['protocol'],
                                           groups=[self.parameters['group']],
                                           cls='enroll'),
                           key=lambda x: (x.client_id, x.id))

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        obj = self.objs[self.next_index - self.start_index]

        # Output: template_id (only provide data when the template_id change)
        if self.outputs['template_id'].isConnected() and \
           self.outputs['template_id'].last_written_data_index < self.next_index:

            end_index = get_client_end_index(self.objs, obj.client_id,
                                             self.next_index,
                                             self.start_index,
                                             self.end_index)

            self.outputs['template_id'].write(
                {
                    'value': np.uint64(obj.client_id)
                },
                end_index
            )

        # Output: client_id (only provide data when the client_id change)
        if self.outputs['client_id'].isConnected() and \
           self.outputs['client_id'].last_written_data_index < self.next_index:

            end_index = get_client_end_index(self.objs, obj.client_id,
                                             self.next_index,
                                             self.start_index,
                                             self.end_index)

            self.outputs['client_id'].write(
                {
                    'value': np.uint64(obj.client_id)
                },
                end_index
            )

        # Output: annotations (provide data at each iteration)
        if self.outputs['annotations'].isConnected():
            annotations = obj.bbx(self.root_folder)

            annotations_list = []
            for i in range(annotations.shape[0]):
                annotations_list.append({
                    'frame_id': np.uint64(annotations[i][0]),
                    'top-left-x': np.int32(annotations[i][1]),
                    'top-left-y': np.int32(annotations[i][2]),
                    'width': np.int32(annotations[i][3]),
                    'height': np.int32(annotations[i][4])
                })

            self.outputs['annotations'].write(
                {
                    'value': annotations_list
                },
                self.next_index
            )

        # Output: file_id (provide data at each iteration)
        if self.outputs['file_id'].isConnected():
            self.outputs['file_id'].write(
                {
                    'value': np.uint64(obj.id)
                },
                self.next_index
            )

        # Output: image (provide data at each iteration)
        if self.outputs['video'].isConnected():
            self.outputs['video'].write(
                {
                    'value': bob.io.base.load(obj.videofile(self.root_folder))
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


class ProbesReal:
    """Outputs:
        - video: "{{ system_user.username }}/array_4d_uint8/1"
        - file_id: "{{ system_user.username }}/uint64/1"
        - annotations: "{{ system_user.username }}/bounding_box_video/1"
        - probe_id: "{{ system_user.username }}/uint64/1"
        - client_id: "{{ system_user.username }}/uint64/1"
        - template_ids: "{{ system_user.username }}/array_1d_uint64/1"

    One "file_id" is associated with a given "video".
    One "annotations" is associated with a given "video".
    One "probe_id" is associated with a given "video".
    Several "video" are associated with a given "client_id".
    Several "client_id" are associated with a given "template_ids".

    --------------- --------------- --------------- ---------------
    |    video    | |    video    | |    video    | |    video    |
    --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- ---------------
    |   file_id   | |   file_id   | |   file_id   | |   file_id   |
    --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- ---------------
    | annotations | | annotations | | annotations | | annotations |
    --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- ---------------
    |  probe_id   | |  probe_id   | |  probe_id   | |  probe_id   |
    --------------- --------------- --------------- ---------------
    ------------------------------- -------------------------------
    |          client_id          | |          client_id          |
    ------------------------------- -------------------------------
    ---------------------------------------------------------------
    |                         template_ids                        |
    ---------------------------------------------------------------

    Note: for this particular database, there is only one "template_ids"
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = root_folder
        self.outputs     = outputs
        self.parameters  = parameters

        # Open the database and load the objects to provide via the outputs
        self.db = bob.db.replay.Database()

        self.template_ids = sorted([c.id for c in self.db.clients() if c.set == parameters['group']])

        self.objs = sorted(self.db.objects(protocol=self.parameters['protocol'],
                                           groups=[self.parameters['group']],
                                           cls='real'),
                           key=lambda x: (x.client_id, x.id))

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        obj = self.objs[self.next_index - self.start_index]

        # Output: template_ids (only provide data once)
        if self.outputs['template_ids'].isConnected() and \
           self.outputs['template_ids'].last_written_data_index < self.next_index:

            self.outputs['template_ids'].write(
                {
                    'value': np.uint64(self.template_ids)
                },
                self.end_index
            )

        # Output: client_id (only provide data when the client_id change)
        if self.outputs['client_id'].isConnected() and \
           self.outputs['client_id'].last_written_data_index < self.next_index:

            end_index = get_client_end_index(self.objs, obj.client_id,
                                             self.next_index,
                                             self.start_index,
                                             self.end_index)

            self.outputs['client_id'].write(
                {
                    'value': np.uint64(obj.client_id)
                },
                end_index
            )

        # Output: annotations (provide data at each iteration)
        if self.outputs['annotations'].isConnected():
            annotations = obj.bbx(self.root_folder)

            annotations_list = []
            for i in range(annotations.shape[0]):
                annotations_list.append({
                    'frame_id': np.uint64(annotations[i][0]),
                    'top-left-x': np.int32(annotations[i][1]),
                    'top-left-y': np.int32(annotations[i][2]),
                    'width': np.int32(annotations[i][3]),
                    'height': np.int32(annotations[i][4])
                })

            self.outputs['annotations'].write(
                {
                    'value': annotations_list
                },
                self.next_index
            )

        # Output: probe_id (provide data at each iteration)
        if self.outputs['probe_id'].isConnected():
            self.outputs['probe_id'].write(
                {
                    'value': np.uint64(obj.id)
                },
                self.next_index
            )

        # Output: file_id (provide data at each iteration)
        if self.outputs['file_id'].isConnected():
            self.outputs['file_id'].write(
                {
                    'value': np.uint64(obj.id)
                },
                self.next_index
            )

        # Output: image (provide data at each iteration)
        if self.outputs['video'].isConnected():
            self.outputs['video'].write(
                {
                    'value': bob.io.base.load(obj.videofile(self.root_folder))
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


class ProbesAttack:
    """Outputs:
        - video: "{{ system_user.username }}/array_4d_uint8/1"
        - file_id: "{{ system_user.username }}/uint64/1"
        - annotations: "{{ system_user.username }}/bounding_box_video/1"
        - probe_id: "{{ system_user.username }}/uint64/1"
        - client_id: "{{ system_user.username }}/uint64/1"
        - attack_support: "{{ system_user.username }}/text/1"
        - template_ids: "{{ system_user.username }}/array_1d_uint64/1"

    One "file_id" is associated with a given "video".
    One "annotations" is associated with a given "video".
    One "probe_id" is associated with a given "video".
    Several "video" are associated with a given "client_id".
    Several "client_id" are associated with a given "template_ids".
    Several "template_ids" are associated with a given "attack_support".

    --------------- --------------- --------------- --------------- --------------- ---------------
    |    video    | |    video    | |    video    | |    video    | |    video    | |    video    |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   | |   file_id   |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    | annotations | | annotations | | annotations | | annotations | | annotations | | annotations |
    --------------- --------------- --------------- --------------- --------------- ---------------
    --------------- --------------- --------------- --------------- --------------- ---------------
    |  probe_id   | |  probe_id   | |  probe_id   | |  probe_id   | |  probe_id   | |  probe_id   |
    --------------- --------------- --------------- --------------- --------------- ---------------
    ------------------------------- ------------------------------- -------------------------------
    |          client_id          | |          client_id          | |          client_id          |
    ------------------------------- ------------------------------- -------------------------------
    --------------------------------------------------------------- -------------------------------
    |                         template_ids                        | |         template_ids        |
    --------------------------------------------------------------- -------------------------------
    -----------------------------------------------------------------------------------------------
    |                                        attack_support                                       |
    -----------------------------------------------------------------------------------------------
    """

    def setup(self, root_folder, outputs, parameters, force_start_index=None,
              force_end_index=None):

        # Initialisations
        self.root_folder = root_folder
        self.outputs     = outputs
        self.parameters  = parameters

        # Open the database and load the objects to provide via the outputs
        self.db = bob.db.replay.Database()

        self.objs = []

        self.objs.extend([ ('fixed', x)
                           for x in sorted(self.db.objects(protocol=parameters['protocol'],
                                                           groups=parameters['group'],
                                                           cls='attack',
                                                           support='fixed'),
                                           key=lambda x: (x.client_id, x.id))
                         ])

        self.objs.extend([ ('hand', x)
                           for x in sorted(self.db.objects(protocol=parameters['protocol'],
                                                           groups=parameters['group'],
                                                           cls='attack',
                                                           support='hand'),
                                           key=lambda x: (x.client_id, x.id))
                         ])

        # Determine the range of indices that must be provided
        self.start_index = force_start_index if force_start_index is not None else 0
        self.end_index = force_end_index if force_end_index is not None else len(self.objs) - 1

        self.objs = self.objs[self.start_index : self.end_index + 1]

        self.next_index = self.start_index

        return True


    def done(self, last_data_index):
        return last_data_index >= self.end_index


    def next(self):
        (attack_support, obj) = self.objs[self.next_index - self.start_index]

        # Output: attack_support (only provide data when the attack_support change)
        if self.outputs['attack_support'].isConnected() and \
           self.outputs['attack_support'].last_written_data_index < self.next_index:

            end_index = get_value_end_index(self.objs, attack_support, 0,
                                            self.next_index,
                                            self.start_index,
                                            self.end_index)

            self.outputs['attack_support'].write(
                {
                    'text': attack_support
                },
                end_index
            )

        # Output: template_ids (only provide data when the template_ids change)
        if self.outputs['template_ids'].isConnected() and \
           self.outputs['template_ids'].last_written_data_index < self.next_index:

            end_index = get_client_end_index(self.objs, obj.client_id,
                                             self.next_index,
                                             self.start_index,
                                             self.end_index)

            self.outputs['template_ids'].write(
                {
                    'value': np.uint64([ obj.client_id ])
                },
                end_index
            )

        # Output: client_id (only provide data when the client_id change)
        if self.outputs['client_id'].isConnected() and \
           self.outputs['client_id'].last_written_data_index < self.next_index:

            end_index = get_client_end_index(self.objs, obj.client_id,
                                             self.next_index,
                                             self.start_index,
                                             self.end_index)

            self.outputs['client_id'].write(
                {
                    'value': np.uint64(obj.client_id)
                },
                end_index
            )

        # Output: annotations (provide data at each iteration)
        if self.outputs['annotations'].isConnected():
            annotations = obj.bbx(self.root_folder)

            annotations_list = []
            for i in range(annotations.shape[0]):
                annotations_list.append({
                    'frame_id': np.uint64(annotations[i][0]),
                    'top-left-x': np.int32(annotations[i][1]),
                    'top-left-y': np.int32(annotations[i][2]),
                    'width': np.int32(annotations[i][3]),
                    'height': np.int32(annotations[i][4])
                })

            self.outputs['annotations'].write(
                {
                    'value': annotations_list
                },
                self.next_index
            )

        # Output: probe_id (provide data at each iteration)
        if self.outputs['probe_id'].isConnected():
            self.outputs['probe_id'].write(
                {
                    'value': np.uint64(obj.id)
                },
                self.next_index
            )

        # Output: file_id (provide data at each iteration)
        if self.outputs['file_id'].isConnected():
            self.outputs['file_id'].write(
                {
                    'value': np.uint64(obj.id)
                },
                self.next_index
            )

        # Output: image (provide data at each iteration)
        if self.outputs['video'].isConnected():
            self.outputs['video'].write(
                {
                    'value': bob.io.base.load(obj.videofile(self.root_folder))
                },
                self.next_index
            )

        # Determine the next data index that must be provided
        self.next_index = 1 + min([ x.last_written_data_index for x in self.outputs
                                                              if x.isConnected() ]
        )

        return True


#----------------------------------------------------------


def setup_tests():
    def mock_load(filename):
        return np.ndarray((5, 3, 10, 20), dtype=np.uint8)

    def mock_bbx(obj, directory):
        return np.array([(0, 1, 2, 3, 4, 5), (1, 10, 20, 30, 40, 50)])

    bob.io.base.load = mock_load
    bob.db.replay.File.bbx = mock_bbx


#----------------------------------------------------------


# Test the behavior of the views (on fake data)
if __name__ == '__main__':

    setup_tests()

    from beat.backend.python.database import DatabaseTester

    DatabaseTester('All', All,
        [
            'class',
            'attack_support',
            'client_id',
            'annotations',
            'file_id',
            'video',
        ],
        parameters=dict(
            protocol="grandtest",
            group="train",
            enroll=False,
        ),
        irregular_outputs=[
            'class',
            'attack_support',
            'client_id',
        ]
    )

    DatabaseTester('Templates', Templates,
        [
            'client_id',
            'template_id',
            'annotations',
            'file_id',
            'video',
        ],
        parameters=dict(
            protocol="grandtest",
            group="devel",
        ),
        irregular_outputs=[
            'client_id',
            'template_id',
        ]
    )

    DatabaseTester('ProbesReal', ProbesReal,
        [
            'template_ids',
            'client_id',
            'probe_id',
            'annotations',
            'file_id',
            'video',
        ],
        parameters=dict(
            protocol="grandtest",
            group="devel",
        ),
        irregular_outputs=[
            'client_id',
        ]
    )

    DatabaseTester('ProbesAttack', ProbesAttack,
        [
            'attack_support',
            'template_ids',
            'client_id',
            'probe_id',
            'annotations',
            'file_id',
            'video',
        ],
        parameters=dict(
            protocol="grandtest",
            group="devel",
        ),
        irregular_outputs=[
            'attack_support',
            'template_ids',
            'client_id',
        ]
    )
